"""Definitions of screw axes with methods for scoring against data."""

from __future__ import annotations

import logging
import math

import numpy as np
from jinja2 import ChoiceLoader, Environment, PackageLoader
from scipy.stats import norm

from scitbx.array_family import flex

from dials.algorithms.symmetry.absences.plots import plot_screw_axes
from dials.util.observer import Observer, Subject, singleton

logger = logging.getLogger("dials.space_group")


@singleton
class ScrewAxisObserver(Observer):
    """Observer to record data used in screw axis analysis."""

    def update(self, screw_axis):
        self.data[screw_axis.name] = {
            "miller_axis_vals": screw_axis.miller_axis_vals,
            "i_over_sigma": screw_axis.i_over_sigma,
            "intensities": screw_axis.intensities,
            "sigmas": screw_axis.sigmas,
            "fourier_space_data": screw_axis.fourier_space_data,
            "axis_repeat": screw_axis.axis_repeat,
        }

    def generate_html_report(self, filename):
        """Generate a html report using the data."""
        screw_axes_graphs = plot_screw_axes(self.data)
        self.data["screw_axes"] = screw_axes_graphs
        loader = ChoiceLoader(
            [
                PackageLoader("dials", "templates"),
                PackageLoader("dials", "static", encoding="utf-8"),
            ]
        )
        env = Environment(loader=loader)
        template = env.get_template("simple_report.html")
        html = template.render(
            page_title="DIALS systematic absences report",
            panel_title="Screw axes analysis",
            panel_id="screw_axes",
            graphs=self.data["screw_axes"],
        )
        with open(filename, "wb") as f:
            f.write(html.encode("utf-8", "xmlcharrefreplace"))


class ScrewAxis(Subject):
    """Definition of a generic screw axis."""

    axis_idx = None  # x=0, y=1, z=2
    axis_repeat = None  # repeat of present reflections e.g =4 for 41, =2 for 42
    name = None

    def __init__(self):
        super().__init__(events=["scored axis"])
        self.equivalent_axes = []
        self.n_refl_used = (0.0, 0.0)
        self.miller_axis_vals = []
        self.i_over_sigma = []
        self.intensities = []
        self.sigmas = []
        self.mean_I_sigma_abs = 0.0
        self.mean_I_sigma = 0.0
        self.mean_I_abs = 0.0
        self.mean_I = 0.0
        self.fourier_space_data = {}

    def add_equivalent_axis(self, equivalent):
        """Add a symmetry equivalent axis."""
        self.equivalent_axes.append(equivalent)

    def select_axial_reflections(self, miller_indices):
        """Select reflections along the screw axis."""
        h, k, l = miller_indices.as_vec3_double().parts()
        if self.axis_idx == 0:
            selection = (k == 0) & (l == 0)
        elif self.axis_idx == 1:
            selection = (h == 0) & (l == 0)
        else:
            selection = (h == 0) & (k == 0)
        return selection

    def get_all_suitable_reflections(self, reflection_table):
        """Select suitable reflections for testing the screw axis."""
        refl = reflection_table
        sel = self.select_axial_reflections(refl["miller_index"])
        miller_idx = refl["miller_index"].select(sel)
        self.miller_axis_vals = miller_idx.as_vec3_double().parts()[self.axis_idx]
        self.intensities = refl["intensity"].select(sel)
        self.sigmas = flex.sqrt(refl["variance"].select(sel))
        self.i_over_sigma = self.intensities / self.sigmas

        if self.equivalent_axes:
            for a in self.equivalent_axes:
                sel = a.select_axial_reflections(refl["miller_index"])
                miller_idx = refl["miller_index"].select(sel)
                intensities = refl["intensity"].select(sel)
                sigmas = flex.sqrt(refl["variance"].select(sel))
                self.i_over_sigma.extend(intensities / sigmas)
                self.miller_axis_vals.extend(
                    miller_idx.as_vec3_double().parts()[a.axis_idx]
                )
                self.intensities.extend(intensities)
                self.sigmas.extend(sigmas)

    @Subject.notify_event(event="scored axis")
    def score_axis(self, reflection_table, significance_level=0.95, method="direct"):
        """Score the axis given a reflection table of data."""
        if method == "direct":
            return self.score_axis_direct(reflection_table, significance_level)
        else:
            return self.score_axis_fourier(reflection_table, significance_level)

    @staticmethod
    def _score_axis_fourier(miller_index, i_over_sigma, axis_repeat):
        """
        Score screw axis based on selected miller indices and I over SigI values

        Parameters
        ----------
        miller_index : np.array
            Vector of miller indices along the chosen axis.
        i_over_sigma : np.array
            Vector of signal to noise ratios with the same shape as miller_indices.
        axis_repeat : int
            The expected periodicity of the screw axis. Must be one of {2, 3, 4, 6}.
        """
        if axis_repeat not in {2, 3, 4, 6}:
            raise ValueError(
                f"Received axis_repeat, {axis_repeat}, but expected an integer in {2, 3, 4, 6}"
            )

        # We must take care to make sure the length of the fourier transformed vector is divisible by 6
        miller_index = miller_index - miller_index.min()
        n = miller_index.max() + 1
        n = 6 * ((n // 6) + 1)
        direct_space = np.zeros(n)
        direct_space[miller_index] = i_over_sigma

        # Fourier transform i over sigma
        fourier_space = np.abs(np.fft.fft(direct_space))

        # Indices for Fourier frequencies which may correspond to screw periodicities
        # These correspond to absences every 0, 2, 3, 4, and 6 reflections.
        screw_idx = [0]
        if axis_repeat == 2:
            screw_idx += [n // 2]
        elif axis_repeat == 3:
            screw_idx += [n // 3, -n // 3]
        elif axis_repeat == 4:
            screw_idx += [n // 4, n // 2, -n // 4]
        else:
            screw_idx += [n // 6, n // 3, n // 2, -n // 3, -n // 6]

        null_idx = np.ones_like(fourier_space, dtype=bool)
        null_idx[screw_idx] = False

        # To determine the probability of a screw axis, use the frequencies which do not
        # correspond to any screw axis periodicity to form a null model. The ask what
        # the probability of the candidate frequency is under the null model.
        # In this case, we will parameterize the null frequencies by a normal distribution.

        mean = fourier_space[null_idx].mean()
        std = fourier_space[null_idx].std()
        p_screw = norm.cdf(fourier_space[n // axis_repeat], loc=mean, scale=std)
        fourier_space_data = {"fourier_space": fourier_space, "n": n}
        return p_screw, fourier_space_data

    def score_axis_fourier(self, reflection_table, significance_level=0.95):
        """Estimate the probability of a screw axis using Fourier analysis."""

        self.get_all_suitable_reflections(reflection_table)
        expected_sel = self.miller_axis_vals.iround() % self.axis_repeat == 0

        expected = self.i_over_sigma.select(expected_sel)
        expected_abs = self.i_over_sigma.select(~expected_sel)
        self.n_refl_used = (expected.size(), expected_abs.size())

        if not expected or not expected_abs:
            return 0.0

        # Log these quantities for reporting.
        self.mean_I_sigma_abs = flex.mean(expected_abs)
        self.mean_I_sigma = flex.mean(expected)
        self.mean_I = flex.mean(self.intensities.select(expected_sel))
        self.mean_I_abs = flex.mean(self.intensities.select(~expected_sel))

        i_over_sigma = np.array(self.i_over_sigma)
        miller_index = np.array(self.miller_axis_vals.iround())
        p_screw, fourier_space_data = self._score_axis_fourier(
            miller_index, i_over_sigma, self.axis_repeat
        )
        # Record the fourier data for reporting.
        self.fourier_space_data = fourier_space_data
        # Zero out the probability if it is less than the significance_level
        if p_screw < significance_level:
            return 0.0

        return p_screw

    def score_axis_direct(self, reflection_table, significance_level=0.95):
        """Score the axis given a reflection table of data."""
        assert significance_level in [0.95, 0.975, 0.99]
        self.get_all_suitable_reflections(reflection_table)

        expected_sel = self.miller_axis_vals.iround() % self.axis_repeat == 0

        expected = self.i_over_sigma.select(expected_sel)
        expected_abs = self.i_over_sigma.select(~expected_sel)
        self.n_refl_used = (expected.size(), expected_abs.size())
        # Limit to best #n reflections to avoid weak at high res - use wilson B?
        if not expected or not expected_abs:
            return 0.0

        # z = (sample mean - population mean) / standard error
        S_E_abs = 1.0  # errors probably correlated so say standard error = 1
        S_E_pres = 1.0  # / expected.size() ** 0.5

        self.mean_I_sigma_abs = flex.mean(expected_abs)
        self.mean_I_sigma = flex.mean(expected)

        self.mean_I = flex.mean(self.intensities.select(expected_sel))
        self.mean_I_abs = flex.mean(self.intensities.select(~expected_sel))

        z_score_absent = self.mean_I_sigma_abs / S_E_abs
        z_score_present = self.mean_I_sigma / S_E_pres

        # get a p-value for z > z_score
        P_absent = 0.5 * (1.0 + math.erf(z_score_absent / (2**0.5)))
        P_present = 0.5 * (1.0 + math.erf(z_score_present / (2**0.5)))

        # sanity check - is most of intensity in 'expected' channel?
        intensity_test = self.mean_I_sigma > (20.0 * self.mean_I_sigma_abs)

        cutoffs = {0.95: 1.645, 0.975: 1.960, 0.99: 2.326}
        cutoff = cutoffs[significance_level]

        if z_score_absent > cutoff and not intensity_test:
            # z > 1.65 in only 5% of cases for normal dist
            # significant nonzero intensity where expected absent.
            return (1.0 - P_absent) * P_present
        elif z_score_absent > cutoff:
            # results appear inconsistent - significant i_over_sigma_abs, but this
            # is still low compared to i_over_sigma_expected
            # try removing the highest absent reflection in case its an outlier
            outlier_msg = (
                """Screw axis %s could only be assigned after removing a suspected outlier
from the expected 'absent' reflections."""
                % self.name
            )
            if expected_abs.size() <= 1:
                logger.info(outlier_msg)
                return P_present
            sel = flex.sort_permutation(expected_abs)
            sorted_exp_abs = expected_abs.select(sel)
            mean_i_sigma_abs = flex.mean(sorted_exp_abs[:-1])
            if (mean_i_sigma_abs / S_E_abs) > cutoff:
                # Still looks like reflections in expected absent
                logger.info(
                    """Test results for %s appear inconsistent (significant nonzero intensity for
'absent' reflections, but majority of intensity in reflection condition).
There may be a set of weak reflections due to pseudosymmetry.""",
                    self.name,
                )
                # Still high intensity of absent, so return as before
                return (1.0 - P_absent) * P_present
            # Looks like there was an outlier, now 'absent' reflections ~ 0.
            self.mean_I_sigma_abs = mean_i_sigma_abs
            self.mean_I_abs = flex.mean(
                self.intensities.select(~expected_sel).select(sel)[:-1]
            )
            if z_score_present > cutoff:
                logger.info(outlier_msg)
                return P_present
            # else in the uncertain case
        elif z_score_present > cutoff:  # evidence with confidence
            return P_present
        logger.info(
            """No evidence to suggest a screw axis for %s, but insufficient
evidence to rule out completely, possibly due to limited data.""",
            self.name,
        )
        return 0.0  # should this be zero or a small number?


class ScrewAxis21c(ScrewAxis):
    """Definition of a 21c screw axis"""

    axis_idx = 2
    axis_repeat = 2
    name = "21c"


class ScrewAxis21b(ScrewAxis):
    """Definition of a 21b screw axis"""

    axis_idx = 1
    axis_repeat = 2
    name = "21b"


class ScrewAxis21a(ScrewAxis):
    """Definition of a 21a screw axis"""

    axis_idx = 0
    axis_repeat = 2
    name = "21a"


class ScrewAxis41c(ScrewAxis):
    """Definition of a 41c screw axis"""

    axis_idx = 2
    axis_repeat = 4
    name = "41c"


class ScrewAxis42c(ScrewAxis):
    """Definition of a 42c screw axis"""

    axis_idx = 2
    axis_repeat = 2
    name = "42c"


class ScrewAxis41b(ScrewAxis):
    """Definition of a 41b screw axis"""

    axis_idx = 1
    axis_repeat = 4
    name = "41b"


class ScrewAxis41a(ScrewAxis):
    """Definition of a 41a screw axis"""

    axis_idx = 0
    axis_repeat = 4
    name = "41a"


class ScrewAxis31c(ScrewAxis):
    """Definition of a 31c screw axis"""

    axis_idx = 2
    axis_repeat = 3
    name = "31c"


class ScrewAxis61c(ScrewAxis):
    """Definition of a 61c screw axis"""

    axis_idx = 2
    axis_repeat = 6
    name = "61c"


class ScrewAxis62c(ScrewAxis):
    """Definition of a 62c screw axis"""

    axis_idx = 2
    axis_repeat = 3
    name = "62c"


class ScrewAxis63c(ScrewAxis):
    """Definition of a 63c screw axis"""

    axis_idx = 2
    axis_repeat = 2
    name = "63c"



from __future__ import annotations

import random

import numpy as np

import scitbx.matrix
import scitbx.random
from cctbx import crystal, sgtbx
from cctbx.sgtbx.subgroups import subgroups
from dxtbx.model import Crystal, Experiment, ExperimentList, Scan

from dials.array_family import flex


def generate_experiments_reflections(
    space_group,
    lattice_group=None,
    unit_cell=None,
    unit_cell_volume=1000,
    seed=0,
    d_min=1,
    sigma=0.1,
    sample_size=100,
    map_to_p1=False,
    twin_fractions=None,
    map_to_minimum=True,
):
    datasets, reindexing_ops = generate_test_data(
        space_group,
        lattice_group=lattice_group,
        unit_cell=unit_cell,
        unit_cell_volume=unit_cell_volume,
        seed=seed,
        d_min=d_min,
        sigma=sigma,
        sample_size=sample_size,
        map_to_p1=map_to_p1,
        twin_fractions=twin_fractions,
        map_to_minimum=map_to_minimum,
    )

    expts = ExperimentList()
    refl_tables = []

    for i, dataset in enumerate(datasets):
        B = scitbx.matrix.sqr(
            dataset.unit_cell().fractionalization_matrix()
        ).transpose()
        expts.append(
            Experiment(
                crystal=Crystal(B, space_group=dataset.space_group(), reciprocal=True),
                scan=Scan(image_range=(0, 180), oscillation=(0.0, 1.0)),
            )
        )
        refl = flex.reflection_table()
        refl["miller_index"] = dataset.indices()
        refl["id"] = flex.int(refl.size(), i)
        refl["d"] = dataset.d_spacings().data()
        refl["intensity.sum.value"] = dataset.data()
        refl["intensity.sum.variance"] = flex.pow2(dataset.sigmas())
        refl.set_flags(flex.bool(len(refl), True), refl.flags.integrated_sum)
        refl_tables.append(refl)
    return expts, refl_tables, reindexing_ops


def generate_test_data(
    space_group,
    lattice_group=None,
    unit_cell=None,
    unit_cell_volume=1000,
    seed=0,
    d_min=1,
    sigma=0.1,
    sample_size=100,
    map_to_p1=False,
    twin_fractions=None,
    map_to_minimum=True,
):
    if seed is not None:
        flex.set_random_seed(seed)
        random.seed(seed)
        np.random.seed(seed)

    assert [unit_cell, lattice_group].count(None) > 0

    sgi = space_group.info()

    if unit_cell is not None:
        cs = crystal.symmetry(unit_cell=unit_cell, space_group_info=sgi)
    elif lattice_group is not None:
        subgrps = subgroups(lattice_group).groups_parent_setting()
        assert space_group in subgrps
        cs = lattice_group.any_compatible_crystal_symmetry(
            volume=unit_cell_volume
        ).customized_copy(space_group_info=sgi)
    else:
        cs = sgi.any_compatible_crystal_symmetry(volume=unit_cell_volume)

    if map_to_minimum:
        cs = cs.minimum_cell()
    intensities = generate_intensities(cs, d_min=d_min)
    intensities.show_summary()

    twin_ops = generate_twin_operators(intensities)
    twin_ops = [sgtbx.change_of_basis_op(op.operator.as_xyz()) for op in twin_ops]

    if twin_fractions is not None:
        assert len(twin_fractions) == len(twin_ops)
        assert len(twin_fractions) == 1, "Only 1 twin component currently supported"
        twin_op = twin_ops[0]
        twin_fraction = twin_fractions[0]
        intensities, intensities_twin = intensities.common_sets(
            intensities.change_basis(twin_op).map_to_asu()
        )
        twinned_miller = intensities.customized_copy(
            data=(1.0 - twin_fraction) * intensities.data()
            + twin_fraction * intensities_twin.data(),
            sigmas=flex.sqrt(
                flex.pow2((1.0 - twin_fraction) * intensities.sigmas())
                + flex.pow2(twin_fraction * intensities_twin.sigmas())
            ),
        )
        intensities = twinned_miller

    cb_ops = twin_ops
    cb_ops.insert(0, sgtbx.change_of_basis_op())

    reindexing_ops = []

    datasets = []
    rand_norm = scitbx.random.normal_distribution(mean=0, sigma=sigma)
    g = scitbx.random.variate(rand_norm)
    for i in range(sample_size):
        cb_op = random.choice(cb_ops)
        reindexing_ops.append(cb_op.as_xyz())
        d = intensities.change_basis(cb_op).customized_copy(
            crystal_symmetry=intensities.crystal_symmetry()
        )

        if map_to_p1:
            cb_op_to_primitive = d.change_of_basis_op_to_primitive_setting()
            d = d.change_basis(cb_op_to_primitive)
            d = d.expand_to_p1()

        d = d.customized_copy(data=d.data() + g(d.size()))
        datasets.append(d)

    return datasets, reindexing_ops


def generate_intensities(crystal_symmetry, anomalous_flag=False, d_min=1):
    from cctbx import miller

    indices = miller.index_generator(
        crystal_symmetry.unit_cell(),
        crystal_symmetry.space_group().type(),
        anomalous_flag,
        d_min,
    ).to_array()
    miller_set = crystal_symmetry.miller_set(indices, anomalous_flag)
    intensities = flex.random_double(indices.size()) * 1000
    miller_array = miller.array(
        miller_set, data=intensities, sigmas=flex.sqrt(intensities)
    ).set_observation_type_xray_intensity()
    return miller_array


def generate_twin_operators(miller_array, verbose=True):
    from mmtbx.scaling.twin_analyses import twin_laws

    TL = twin_laws(miller_array=miller_array)
    return TL.operators



"""Definitions of plots for systematic absences."""


from __future__ import annotations


def color_axis_data(name, miller_axis_vals):
    """Generate a sequence for coloring the datapoints for a screw axis."""
    if name.startswith("41"):
        colors = [1 if m % 4 == 0 else 0 for m in miller_axis_vals]
    elif name.startswith("21") or name.startswith("42") or name.startswith("63"):
        colors = [1 if m % 2 == 0 else 0 for m in miller_axis_vals]
    elif name.startswith("31") or name.startswith("62"):
        colors = [1 if m % 3 == 0 else 0 for m in miller_axis_vals]
    elif name.startswith("61"):
        colors = [1 if m % 6 == 0 else 0 for m in miller_axis_vals]
    return colors


def plot_screw_axes(screw_axes_data):
    """Generate scatter plot data for screw axes."""
    d = {}
    for name, data in screw_axes_data.items():
        d.update(
            {
                "plot_"
                + name: {
                    "data": [
                        {
                            "x": list(data["miller_axis_vals"]),
                            "y": list(data["i_over_sigma"]),
                            "type": "scatter",
                            "name": name,
                            "xaxis": "x",
                            "yaxis": "y",
                            "mode": "markers",
                            "marker": {
                                "color": color_axis_data(
                                    name, list(data["miller_axis_vals"])
                                ),
                                "colorscale": "Viridis",
                            },
                        }
                    ],
                    "layout": {
                        "title": f"I (merged) / σ (merged) along axis {name}",
                        "xaxis": {
                            "domain": [0, 1],
                            "anchor": "y",
                            "title": "index along axis",
                        },
                        "yaxis": {"domain": [0, 1], "anchor": "x", "title": "I/σ(I)"},
                    },
                }
            }
        )
        if data["fourier_space_data"]:
            # make frequency plots
            # need to add vertical lines to indicate axis repeats
            y_min = min(data["fourier_space_data"]["fourier_space"])
            y_max = max(data["fourier_space_data"]["fourier_space"])
            n = data["fourier_space_data"]["n"]
            xtickvals = [float(n // 2)]
            xticktext = ["1/2", "1/3", "2/3", "1/4", "3/4", "1/6", "5/6"]
            for i in [3, 4, 6]:
                xloc = float(n // i)
                xloc2 = n - xloc
                xtickvals.extend([xloc, xloc2])

            plot = {
                "frequencies_plot_"
                + name: {
                    "data": [
                        {
                            "x": list(
                                range(len(data["fourier_space_data"]["fourier_space"]))
                            ),
                            "y": list(data["fourier_space_data"]["fourier_space"]),
                            "type": "scatter",
                            "name": "Fourier amplitudes",
                            "xaxis": "x",
                            "yaxis": "y",
                            "mode": "lines",
                        },
                    ],
                    "layout": {
                        "title": "Fourier amplitudes for axis " + name,
                        "xaxis": {
                            "title": "frequency",
                            "tickvals": xtickvals,
                            "ticktext": xticktext,
                        },
                        "yaxis": {"title": "Amplitude", "domain": [y_min, y_max]},
                    },
                    "help": """
This plot shows the amplitudes of the frequency spectrum of I/sigma, calculated via
a discrete fourier transform. The frequencies corresponding to systematic absences
are indicated by vertical lines.""",
                }
            }
            for i, dash in zip([2, 3, 4, 6], ["dash", "dot", "dashdot", "solid"]):
                xloc = float(n // i)
                plot["frequencies_plot_" + name]["data"].append(
                    {
                        "x": [xloc, xloc],
                        "y": [y_min, y_max],
                        "type": "scatter",
                        "xaxis": "x",
                        "yaxis": "y",
                        "mode": "lines",
                        "line": {
                            "color": "black" if i == data["axis_repeat"] else "grey",
                            "dash": dash,
                        },
                        "name": f"{i}-fold repeat",
                    }
                )
                if i == 2:
                    continue  # don't plot same thing twice
                xloc2 = float(n - xloc)
                plot["frequencies_plot_" + name]["data"].append(
                    {
                        "x": [xloc2, xloc2],
                        "y": [y_min, y_max],
                        "type": "scatter",
                        "xaxis": "x",
                        "yaxis": "y",
                        "mode": "lines",
                        "line": {
                            "color": "black" if i == data["axis_repeat"] else "grey",
                            "dash": dash,
                        },
                        "name": None,
                        "showlegend": False,
                    }
                )
            d.update(plot)

        d.update(
            {
                "intensities_plot_"
                + name: {
                    "data": [
                        {
                            "x": list(data["miller_axis_vals"]),
                            "y": list(data["intensities"]),
                            "type": "scatter",
                            "name": "intensity",
                            "xaxis": "x",
                            "yaxis": "y",
                            "mode": "markers",
                            "marker": {
                                "color": color_axis_data(
                                    name, list(data["miller_axis_vals"])
                                ),
                                "colorscale": "Viridis",
                            },
                        },
                        {
                            "x": list(data["miller_axis_vals"]),
                            "y": list(data["sigmas"]),
                            "type": "scatter",
                            "name": "sigma",
                            "xaxis": "x",
                            "yaxis": "y",
                            "mode": "markers",
                        },
                    ],
                    "layout": {
                        "title": f"I, σ (merged) along axis {name}",
                        "xaxis": {
                            "domain": [0, 1],
                            "anchor": "y",
                            "title": "index along axis",
                        },
                        "yaxis": {"domain": [0, 1], "anchor": "x", "title": "I, σ"},
                    },
                }
            }
        )
    return d



from __future__ import annotations

import dials_algorithms_spot_prediction_ext
from dials_algorithms_spot_prediction_ext import (
    IndexGenerator,
    NaveStillsReflectionPredictor,
    PixelLabeller,
    PixelToMillerIndex,
    ReekeIndexGenerator,
    RotationAngles,
    ScanStaticRayPredictor,
    ScanVaryingRayPredictor,
    SphericalRelpStillsReflectionPredictor,
    StillsDeltaPsiReflectionPredictor,
    StillsRayPredictor,
    ray_intersection,
)

__all__ = [
    "IndexGenerator",
    "NaveStillsReflectionPredictor",
    "PixelLabeller",
    "PixelToMillerIndex",
    "ray_intersection",
    "ReekeIndexGenerator",
    "RotationAngles",
    "ScanStaticRayPredictor",
    "ScanStaticReflectionPredictor",
    "ScanVaryingRayPredictor",
    "ScanVaryingReflectionPredictor",
    "SphericalRelpStillsReflectionPredictor",
    "StillsDeltaPsiReflectionPredictor",
    "StillsRayPredictor",
    "StillsReflectionPredictor",
]


def ScanStaticReflectionPredictor(experiment, dmin=None, margin=1, padding=0, **kwargs):
    """
    A constructor for the reflection predictor.

    :param experiment: The experiment to predict for
    :param dmin: The maximum resolution to predict to
    :param margin: The margin for prediction
    :return: The spot predictor
    """

    # Get dmin if it is not set
    if dmin is None:
        dmin = experiment.detector.get_max_resolution(experiment.beam.get_s0())

    # Only remove certain systematic absences
    space_group = experiment.crystal.get_space_group()
    space_group = space_group.build_derived_patterson_group()

    # Create the reflection predictor
    return dials_algorithms_spot_prediction_ext.ScanStaticReflectionPredictor(
        experiment.beam,
        experiment.detector,
        experiment.goniometer,
        experiment.scan,
        experiment.crystal.get_unit_cell(),
        space_group.type(),
        dmin,
        margin,
        padding,
    )


def ScanVaryingReflectionPredictor(
    experiment, dmin=None, margin=1, padding=0, **kwargs
):
    """
    A constructor for the reflection predictor.

    :param experiment: The experiment to predict for
    :param dmin: The maximum resolution to predict to
    :param margin: The margin for prediction
    :return: The spot predictor
    """

    # Get dmin if it is not set
    if dmin is None:
        dmin = experiment.detector.get_max_resolution(experiment.beam.get_s0())

    # Only remove certain systematic absences
    space_group = experiment.crystal.get_space_group()
    space_group = space_group.build_derived_patterson_group()

    # Create the reflection predictor
    return dials_algorithms_spot_prediction_ext.ScanVaryingReflectionPredictor(
        experiment.beam,
        experiment.detector,
        experiment.goniometer,
        experiment.scan,
        space_group.type(),
        dmin,
        margin,
        padding,
    )


def StillsReflectionPredictor(experiment, dmin=None, spherical_relp=False, **kwargs):
    """
    A factory function for the reflection predictor.

    :param experiment: The experiment to predict for
    :param dmin: The maximum resolution to predict to
    :param spherical_relp: Whether to use the spherical relp prediction model
    :return: The spot predictor
    """

    # FIXME Selection of reflection predictor type is ugly. What is a better
    # way here? Should it be based entirely on the existence of certain types
    # of profile model within the experiment?

    # Get dmin if it is not set
    if dmin is None:
        dmin = experiment.detector.get_max_resolution(experiment.beam.get_s0())

    if spherical_relp:
        return SphericalRelpStillsReflectionPredictor(
            experiment.beam,
            experiment.detector,
            experiment.crystal.get_A(),
            experiment.crystal.get_unit_cell(),
            experiment.crystal.get_space_group().type(),
            dmin,
        )

    # Create the reflection predictor
    try:
        if (
            experiment.crystal.get_half_mosaicity_deg() is not None
            and experiment.crystal.get_domain_size_ang() is not None
        ):
            return NaveStillsReflectionPredictor(
                experiment.beam,
                experiment.detector,
                experiment.crystal.get_A(),
                experiment.crystal.get_unit_cell(),
                experiment.crystal.get_space_group().type(),
                dmin,
                experiment.crystal.get_half_mosaicity_deg(),
                experiment.crystal.get_domain_size_ang(),
            )
    except AttributeError:
        pass

    return StillsDeltaPsiReflectionPredictor(
        experiment.beam,
        experiment.detector,
        experiment.crystal.get_A(),
        experiment.crystal.get_unit_cell(),
        experiment.crystal.get_space_group().type(),
        dmin,
    )



"""
Tools for analysis of missing reflections.
"""

from __future__ import annotations

from annlib_ext import AnnAdaptor

import cctbx.miller
from boost_adaptbx import graph
from boost_adaptbx.graph import connected_component_algorithm as cca

from dials.array_family import flex


def connected_components(
    miller_array: cctbx.miller.array,
) -> [{}]:
    """
    Identify connected regions of missing reflections in the asymmetric unit.

    This is achieved by first generating the complete set of possible miller indices,
    then performing connected components analysis on a graph of nearest neighbours in
    the list of missing reflections.

    Args:
        miller_array:  The input list of reflections.

    Returns:
        The list of miller sets for each connected region of missing reflections. The
        first item in the list will be the complete set of all possible miller indices.
    """

    # Map to primitive setting for centred cells, otherwise true missing reflections
    # won't be identified as connected as a result of being separated by systematically
    # absent reflections.
    cb_op_to_primitive = miller_array.change_of_basis_op_to_primitive_setting()
    miller_array = miller_array.change_basis(cb_op_to_primitive)

    # First generate the missing_set of reflections. We want the full sphere of missing
    # reflections to allow us to find connected regions that cross the boundary of the
    # asu.
    unique = miller_array.unique_under_symmetry().map_to_asu()
    unique = unique.generate_bijvoet_mates()
    complete_set = unique.complete_set()
    missing_set = complete_set.lone_set(unique)
    missing_set = missing_set.expand_to_p1().customized_copy(
        crystal_symmetry=missing_set.crystal_symmetry()
    )

    if missing_set.size() == 0:
        return complete_set, []

    # Now find the nearest neighbours.
    mi = missing_set.indices().as_vec3_double().as_double()
    k = 6
    ann = AnnAdaptor(data=mi, dim=3, k=k)
    ann.query(mi)

    # Construct the graph of connected missing reflections
    g = graph.adjacency_list(
        graph_type="undirected",
        vertex_type="vector",
        edge_type="set",
    )
    distance_cutoff = 2**0.5
    for i in range(missing_set.size()):
        ik = i * k
        for i_ann in range(k):
            if ann.distances[ik + i_ann] <= distance_cutoff:
                j = ann.nn[ik + i_ann]
                g.add_edge(i, j)

    # Now do the connected components analysis, filtering out lone missing reflections
    components = [c for c in cca.connected_components(graph=g) if len(c) > 1]

    # Determine the unique miller indices for each component within the asu
    unique_mi = []
    unique_ms = []
    for i, c in enumerate(components):
        ms = (
            missing_set.select(flex.size_t(list(c)))
            .customized_copy(crystal_symmetry=miller_array)
            .as_non_anomalous_set()
            .map_to_asu()
        )
        ms = ms.unique_under_symmetry()
        mi = set(ms.indices())
        if mi not in unique_mi:
            unique_ms.append(ms)
            unique_mi.append(mi)

    # Sort connected regions by size
    unique_ms = sorted(unique_ms, key=lambda ms: ms.size(), reverse=True)

    # Map indices back to input setting
    cb_op_primitive_inp = cb_op_to_primitive.inverse()
    return (
        unique.as_non_anomalous_set().complete_set().change_basis(cb_op_primitive_inp),
        [ms.change_basis(cb_op_primitive_inp) for ms in unique_ms],
    )



from __future__ import annotations

import procrunner


def test(dials_data, tmp_path):
    input_filename = (
        dials_data("centroid_test_data", pathlib=True) / "imported_experiments.json"
    )

    result = procrunner.run(
        ["dials.estimate_gain", f"input.experiments={input_filename}"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert b"Estimated gain: 1.0" in result.stdout



from __future__ import annotations

import pathlib

import procrunner

from dxtbx.model.experiment_list import ExperimentListFactory


def test_experiments(dials_data, tmp_path):
    input_filename = (
        dials_data("centroid_test_data", pathlib=True) / "imported_experiments.json"
    )
    mask_filename = (
        dials_data("centroid_test_data", pathlib=True) / "lookup_mask.pickle"
    )
    output_filename = tmp_path / "output.expt"

    result = procrunner.run(
        [
            "dials.apply_mask",
            f"input.experiments={input_filename}",
            f"input.mask={mask_filename}",
            f"output.experiments={output_filename}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    experiments = ExperimentListFactory.from_json_file(output_filename)

    assert len(experiments) == 1
    imageset = experiments[0].imageset
    assert pathlib.Path(imageset.external_lookup.mask.filename) == mask_filename



from __future__ import annotations

import procrunner

from dxtbx.serialize import load

from dials.array_family import flex


def test_for_preservation_of_identifiers_in_dials_processing(dials_data, tmp_path):
    """Run the dials processing workflow, checking for preservation of identifiers.

    This is just a simple case. The individual programs that are expected to
    change the identifiers are tested separately, this is to check that the
    other programs maintain the identifiers through processing.
    """

    imported = tmp_path / "imported.expt"
    strong_refl = tmp_path / "strong.refl"
    indexed_expt = tmp_path / "indexed.expt"
    indexed_refl = tmp_path / "indexed.refl"
    bravais_expt = tmp_path / "bravais_setting_9.expt"
    reindexed_expt = tmp_path / "reindexed.expt"
    reindexed_refl = tmp_path / "reindexed.refl"
    refined_expt = tmp_path / "refined.expt"
    refined_refl = tmp_path / "refined.refl"
    integrated_expt = tmp_path / "integrated.expt"
    integrated_refl = tmp_path / "integrated.refl"
    symmetrized_expt = tmp_path / "symmetrized.expt"
    symmetrized_refl = tmp_path / "symmetrized.refl"
    scaled_expt = tmp_path / "scaled.expt"
    scaled_refl = tmp_path / "scaled.refl"
    tt_expt = tmp_path / "tt.expt"

    # First import - should set a unique id.
    image_files = dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    result = procrunner.run(
        ["dials.import", f"output.experiments={imported.name}"] + sorted(image_files),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert imported.is_file()
    experiments = load.experiment_list(imported)
    import_expt_id = experiments[0].identifier
    assert import_expt_id != ""

    # Now find spots.
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            imported,
            f"output.reflections={strong_refl.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert strong_refl.is_file()
    reflections = flex.reflection_table.from_file(strong_refl)
    assert dict(reflections.experiment_identifiers()) == {0: import_expt_id}

    # Now index
    result = procrunner.run(
        [
            "dials.index",
            strong_refl,
            imported,
            f"output.reflections={indexed_refl.name}",
            f"output.experiments={indexed_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert indexed_expt.is_file()
    assert indexed_refl.is_file()
    experiments = load.experiment_list(indexed_expt)
    reflections = flex.reflection_table.from_file(indexed_refl)

    indexed_expt_id = experiments[0].identifier
    assert indexed_expt_id
    assert list(experiments.identifiers()) == [indexed_expt_id]
    assert dict(reflections.experiment_identifiers()) == {0: indexed_expt_id}

    # Now refine bravais setting
    result = procrunner.run(
        ["dials.refine_bravais_settings", indexed_refl, indexed_expt],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert bravais_expt.is_file()
    experiments = load.experiment_list(bravais_expt)
    assert experiments[0].identifier == indexed_expt_id

    # Now reindex
    result = procrunner.run(
        [
            "dials.reindex",
            indexed_refl,
            indexed_expt,
            "change_of_basis_op=b,c,a",
            f"output.reflections={reindexed_refl.name}",
            f"output.experiments={reindexed_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert reindexed_expt.is_file()
    assert reindexed_refl.is_file()
    experiments = load.experiment_list(reindexed_expt)
    reflections = flex.reflection_table.from_file(reindexed_refl)

    assert list(experiments.identifiers()) == [indexed_expt_id]
    assert dict(reflections.experiment_identifiers()) == {0: indexed_expt_id}

    # Now refine
    result = procrunner.run(
        [
            "dials.refine",
            reindexed_refl,
            reindexed_expt,
            f"output.reflections={refined_refl.name}",
            f"output.experiments={refined_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert refined_expt.is_file()
    assert refined_refl.is_file()
    experiments = load.experiment_list(refined_expt)
    reflections = flex.reflection_table.from_file(refined_refl)

    assert list(experiments.identifiers()) == [indexed_expt_id]
    assert dict(reflections.experiment_identifiers()) == {0: indexed_expt_id}

    # Now integrate
    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            refined_refl,
            refined_expt,
            f"output.reflections={integrated_refl.name}",
            f"output.experiments={integrated_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert integrated_expt.is_file()
    assert integrated_refl.is_file()
    experiments = load.experiment_list(integrated_expt)
    reflections = flex.reflection_table.from_file(integrated_refl)

    assert list(experiments.identifiers()) == [indexed_expt_id]
    assert dict(reflections.experiment_identifiers()) == {0: indexed_expt_id}

    # Now run cosym (symmetry fails due to small amount of data)
    result = procrunner.run(
        [
            "dials.symmetry",
            integrated_refl,
            integrated_expt,
            f"output.reflections={symmetrized_refl.name}",
            f"output.experiments={symmetrized_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert symmetrized_expt.is_file()
    assert symmetrized_refl.is_file()
    experiments = load.experiment_list(symmetrized_expt)
    reflections = flex.reflection_table.from_file(symmetrized_refl)

    assert list(experiments.identifiers()) == [indexed_expt_id]
    assert dict(reflections.experiment_identifiers()) == {0: indexed_expt_id}

    # Now scale
    result = procrunner.run(
        [
            "dials.scale",
            symmetrized_refl,
            symmetrized_expt,
            f"output.reflections={scaled_refl.name}",
            f"output.experiments={scaled_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert scaled_expt.is_file()
    assert scaled_refl.is_file()
    experiments = load.experiment_list(scaled_expt)
    reflections = flex.reflection_table.from_file(scaled_refl)

    assert list(experiments.identifiers()) == [indexed_expt_id]
    assert dict(reflections.experiment_identifiers()) == {0: indexed_expt_id}

    # Now do two-theta refine
    result = procrunner.run(
        [
            "dials.two_theta_refine",
            scaled_refl,
            scaled_expt,
            f"output.experiments={tt_expt.name}",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert tt_expt.is_file()
    experiments = load.experiment_list(tt_expt)

    assert list(experiments.identifiers()) == [indexed_expt_id]



from __future__ import annotations

from scitbx import matrix

from dials.algorithms.refinement.parameterisation.goniometer_parameters import (
    GoniometerMixin,
)
from dials.algorithms.refinement.parameterisation.scan_varying_model_parameters import (
    GaussianSmoother,
    ScanVaryingModelParameterisation,
    ScanVaryingParameterSet,
)


class ScanVaryingGoniometerParameterisation(
    ScanVaryingModelParameterisation, GoniometerMixin
):
    """A scan-varying parameterisation for the setting rotation of a goniometer
    with angles expressed in mrad."""

    def __init__(
        self, goniometer, t_range, num_intervals, beam=None, experiment_ids=None
    ):

        if experiment_ids is None:
            experiment_ids = [0]

        # The state of a scan varying goniometer parameterisation is a matrix
        # '[S](t)', expressed as a function of image number 't'
        # in a sequential scan.
        #
        # The initial state is a snapshot of the setting matrix
        # at the point of initialisation '[S0]', which is independent of
        # image number.
        #
        # Future states are composed by rotations around two axes orthogonal to the
        # initial spindle axis direction.
        #
        # [S](t) = [G2](t)[G1](t)[S0]

        # Set up the smoother
        smoother = GaussianSmoother(t_range, num_intervals)
        nv = smoother.num_values()

        # Set up the initial state
        e_lab = matrix.col(goniometer.get_rotation_axis())
        istate = matrix.sqr(goniometer.get_setting_rotation())
        self._S_at_t = istate

        # Factory function to provide to _build_p_list
        def parameter_type(value, axis, ptype, name):
            return ScanVaryingParameterSet(value, nv, axis, ptype, name)

        # Build the parameter list
        p_list = self._build_p_list(e_lab, beam, parameter_type=parameter_type)

        # Set up the base class
        ScanVaryingModelParameterisation.__init__(
            self, goniometer, istate, p_list, smoother, experiment_ids=experiment_ids
        )

        return

    def compose(self, t):
        """calculate state and derivatives for model at image number t"""

        # Extract setting matrix from the initial state
        iS0 = self._initial_state

        # extract parameter sets from the internal list
        gamma1_set, gamma2_set = self._param

        # extract angles and other data at time t using the smoother
        gamma1, gamma1_weights, gamma1_sumweights = self._smoother.value_weight(
            t, gamma1_set
        )
        gamma2, gamma2_weights, gamma2_sumweights = self._smoother.value_weight(
            t, gamma2_set
        )

        # calculate derivatives of angles wrt underlying parameters.
        dgamma1_dp = gamma1_weights * (1.0 / gamma1_sumweights)
        dgamma2_dp = gamma2_weights * (1.0 / gamma2_sumweights)

        self._S_at_t, dS_dval = self._compose_core(
            iS0,
            gamma1,
            gamma2,
            gamma1_axis=gamma1_set.axis,
            gamma2_axis=gamma2_set.axis,
        )

        # calculate derivatives of state wrt underlying smoother parameters
        dS_dp1 = [None] * dgamma1_dp.size
        for (i, v) in dgamma1_dp:
            dS_dp1[i] = dS_dval[0] * v
        dS_dp2 = [None] * dgamma2_dp.size
        for (i, v) in dgamma2_dp:
            dS_dp2[i] = dS_dval[1] * v

        # store derivatives as list-of-lists
        self._dstate_dp = [dS_dp1, dS_dp2]

        return

    def get_state(self):
        """Return setting matrix [S] at image number t"""

        # only a single goniometer is parameterised here, so no multi_state_elt
        # argument is allowed

        return self._S_at_t



from __future__ import annotations

from libtbx import phil
from scitbx import matrix
from scitbx.array_family import flex

from .strategy import Strategy

fft1d_phil_str = """\
characteristic_grid = None
    .help = Sampling frequency in radians. See Steller 1997. If None, \
            determine a grid sampling automatically using the input \
            reflections, using at most 0.029 radians.
    .type = float(value_min=0)
"""


class FFT1D(Strategy):
    """
    Basis vector search using 1D FFTs in reciprocal space.

    A set of dimensionless radial unit vectors, typically ~7000 in total, is chosen
    so that they are roughly evenly spaced in solid angle over a hemisphere. The
    reciprocal space displacements of the measured spot centroids are then projected
    onto each of these radial vectors in turn (that is, we calculate the scalar
    product of each displacement with each unit vector). A 1D FFT of the linear
    density of projected spot positions is performed along each direction.
    Aggregating the results of all the transforms, the three shortest non-collinear
    wave vectors with the greatest spectral weight correspond to the basis vectors of
    the direct lattice.

    See:
        Steller, I., Bolotovsky, R. & Rossmann, M. G. (1997). J. Appl. Cryst. 30, 1036-1040.
        Sauter, N. K., Grosse-Kunstleve, R. W. & Adams, P. D. (2004). J. Appl. Cryst. 37, 399-409.
    """

    phil_help = (
        "Search for the basis vectors of the direct lattice by performing a series of "
        "1D FFTs along various directions in reciprocal space. This has a lower "
        "memory requirement than a single 3D FFT (the fft3d method). This method may "
        "also be more appropriate than a 3D FFT if the reflections are from narrow "
        "wedges of rotation data or from stills data."
    )

    phil_scope = phil.parse(fft1d_phil_str)

    def __init__(self, max_cell, params=None, *args, **kwargs):
        """Construct an FFT1D object.

        Args:
            max_cell (float): An estimate of the maximum cell dimension of the primitive
                cell.
            characteristic_grid (float): Sampling frequency in radians. See Steller 1997.
                If None, determine a grid sampling automatically using the input
                reflections, using at most 0.029 radians.
        """
        super().__init__(max_cell, params=params, *args, **kwargs)

    def find_basis_vectors(self, reciprocal_lattice_vectors):
        """Find a list of likely basis vectors.

        Args:
            reciprocal_lattice_vectors (scitbx.array_family.flex.vec3_double):
                The list of reciprocal lattice vectors to search for periodicity.

        Returns:
            A tuple containing the list of basis vectors and a flex.bool array
            identifying which reflections were used in indexing.
        """
        import iotbx.phil
        from rstbx.phil.phil_preferences import indexing_api_defs

        used_in_indexing = flex.bool(reciprocal_lattice_vectors.size(), True)

        hardcoded_phil = iotbx.phil.parse(input_string=indexing_api_defs).extract()

        # Spot_positions: Centroid positions for spotfinder spots, in pixels
        # Return value: Corrected for parallax, converted to mm

        # derive a max_cell from mm spots
        # derive a grid sampling from spots

        from rstbx.indexing_api.lattice import DPS_primitive_lattice

        # max_cell: max possible cell in Angstroms; set to None, determine from data
        # recommended_grid_sampling_rad: grid sampling in radians; guess for now

        DPS = DPS_primitive_lattice(
            max_cell=self._max_cell,
            recommended_grid_sampling_rad=self._params.characteristic_grid,
            horizon_phil=hardcoded_phil,
        )

        # transform input into what Nick needs
        # i.e., construct a flex.vec3 double consisting of mm spots, phi in degrees
        DPS.index(reciprocal_space_vectors=reciprocal_lattice_vectors)
        solutions = DPS.getSolutions()
        candidate_basis_vectors = [matrix.col(s.bvec()) for s in solutions]
        return candidate_basis_vectors, used_in_indexing



from __future__ import annotations

import math

import numpy as np
import pytest

from dials.algorithms.integration.fit import ProfileFitter
from dials.array_family import flex


def evaluate_gaussian(x, a, x0, sx):
    assert len(x) == len(x0)
    assert len(x) == len(sx)

    g = 0.0
    for xi, x0i, sxi in zip(x, x0, sx):
        g += (xi - x0i) ** 2 / (2.0 * sxi**2)

    return a * math.exp(-g)


def gaussian(size, a, x0, sx):
    result = flex.double(flex.grid(size))

    index = [0] * len(size)
    while True:
        result[index] = evaluate_gaussian(index, a, x0, sx)
        for j in range(len(size)):
            index[j] += 1
            if index[j] < size[j]:
                break
            index[j] = 0
            if j == len(size) - 1:
                return result


def add_poisson_noise(x):
    s = x.accessor()
    y = flex.double(np.random.poisson(xx) for xx in x)
    y.reshape(s)
    return y


def generate_3_profiles():
    p1 = gaussian((40, 9, 9), 1, (10.5, 4, 4), (2, 2, 2))
    p2 = gaussian((40, 9, 9), 1, (20.5, 4, 4), (2, 2, 2))
    p3 = gaussian((40, 9, 9), 1, (30.5, 4, 4), (2, 2, 2))
    p1 = p1 / flex.sum(p1)
    p2 = p2 / flex.sum(p2)
    p3 = p3 / flex.sum(p3)
    p1.reshape(flex.grid(1, 40, 9, 9))
    p2.reshape(flex.grid(1, 40, 9, 9))
    p3.reshape(flex.grid(1, 40, 9, 9))
    p = flex.double(flex.grid(3, 40, 9, 9))
    p[0:1, :, :, :] = p1
    p[1:2, :, :, :] = p2
    p[2:3, :, :, :] = p3
    return p


def generate_7_profiles():
    p1 = gaussian((40, 40, 40), 1, (10.5, 20.5, 20.5), (2, 2, 2))
    p2 = gaussian((40, 40, 40), 1, (20.5, 20.5, 20.5), (2, 2, 2))
    p3 = gaussian((40, 40, 40), 1, (30.5, 20.5, 20.5), (2, 2, 2))
    p4 = gaussian((40, 40, 40), 1, (20.5, 10.5, 20.5), (2, 2, 2))
    p5 = gaussian((40, 40, 40), 1, (20.5, 30.5, 20.5), (2, 2, 2))
    p6 = gaussian((40, 40, 40), 1, (20.5, 20.5, 10.5), (2, 2, 2))
    p7 = gaussian((40, 40, 40), 1, (20.5, 20.5, 30.5), (2, 2, 2))
    p1 = p1 / flex.sum(p1)
    p2 = p2 / flex.sum(p2)
    p3 = p3 / flex.sum(p3)
    p4 = p4 / flex.sum(p4)
    p5 = p5 / flex.sum(p5)
    p6 = p6 / flex.sum(p6)
    p7 = p7 / flex.sum(p7)
    p1.reshape(flex.grid(1, 40, 40, 40))
    p2.reshape(flex.grid(1, 40, 40, 40))
    p3.reshape(flex.grid(1, 40, 40, 40))
    p4.reshape(flex.grid(1, 40, 40, 40))
    p5.reshape(flex.grid(1, 40, 40, 40))
    p6.reshape(flex.grid(1, 40, 40, 40))
    p7.reshape(flex.grid(1, 40, 40, 40))
    p = flex.double(flex.grid(7, 40, 40, 40))
    p[0:1, :, :, :] = p1
    p[1:2, :, :, :] = p2
    p[2:3, :, :, :] = p3
    p[3:4, :, :, :] = p4
    p[4:5, :, :, :] = p5
    p[5:6, :, :, :] = p6
    p[6:7, :, :, :] = p7
    return p


def test_zero():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c = flex.double(flex.grid(9, 9, 9), 0)
    b = flex.double(flex.grid(9, 9, 9), 0)
    m = flex.bool(flex.grid(9, 9, 9), True)

    # Fit
    fit = ProfileFitter(c, b, m, p)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    # Test intensity is the same
    eps = 1e-3
    assert intensity[0] == pytest.approx(flex.sum(c), abs=eps)
    assert intensity[0] == pytest.approx(V[0], abs=eps)


def test_identical():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c = p.deep_copy()
    b = flex.double(flex.grid(9, 9, 9), 0)
    m = flex.bool(flex.grid(9, 9, 9), True)

    # Fit
    fit = ProfileFitter(c, b, m, p)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    # Test intensity is the same
    eps = 1e-3
    assert intensity[0] == pytest.approx(flex.sum(c), abs=eps)
    assert intensity[0] == pytest.approx(V[0], abs=eps)


def test_with_no_background():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c = add_poisson_noise(100 * p)
    b = flex.double(flex.grid(9, 9, 9), 0)
    m = flex.bool(flex.grid(9, 9, 9), True)

    # Fit
    fit = ProfileFitter(c, b, m, p)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    # Test intensity is the same
    eps = 1e-3
    assert intensity[0] == pytest.approx(flex.sum(c), abs=eps)
    assert intensity[0] == pytest.approx(V[0], abs=eps)


def test_mask():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c = add_poisson_noise(100 * p)
    b = flex.double(flex.grid(9, 9, 9), 0)
    m = flex.bool(flex.grid(9, 9, 9), True)
    m[4, 4, 4] = False

    # Fit
    fit = ProfileFitter(c, b, m, p)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    # Test intensity is approximately the same
    assert intensity[0] == pytest.approx(flex.sum(c), rel=0.01)
    assert intensity[0] == pytest.approx(V[0], abs=1e-3)


def test_with_flat_background():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c0 = add_poisson_noise(100 * p)
    b = flex.double(flex.grid(9, 9, 9), 10)
    m = flex.bool(flex.grid(9, 9, 9), True)
    b0 = add_poisson_noise(b)
    c = c0 + b0

    # Fit
    fit = ProfileFitter(c, b, m, p)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    Iknown = 201.67417836585147
    Vknown = 7491.6743173001205

    # Test intensity is the same
    eps = 1e-3
    assert intensity[0] == pytest.approx(Iknown, abs=eps)
    assert V[0] == pytest.approx(Vknown, abs=eps)


def test_identical_partial():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c = p.deep_copy()
    b = flex.double(flex.grid(9, 9, 9), 0)
    m = flex.bool(flex.grid(9, 9, 9), True)

    # Get the partial profiles
    pp = p[0:5, :, :]
    mp = m[0:5, :, :]
    cp = c[0:5, :, :]
    bp = b[0:5, :, :]

    # Fit
    fit = ProfileFitter(cp, bp, mp, pp)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    # Test intensity is the same
    eps = 1e-7
    assert intensity[0] == pytest.approx(flex.sum(p), abs=eps)
    assert V[0] == pytest.approx(flex.sum(p), abs=eps)


def test_with_no_background_partial():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c = add_poisson_noise(100 * p)
    b = flex.double(flex.grid(9, 9, 9), 0)
    m = flex.bool(flex.grid(9, 9, 9), True)

    # Get the partial profiles
    pp = p[0:5, :, :]
    mp = m[0:5, :, :]
    cp = c[0:5, :, :]
    bp = b[0:5, :, :]

    # Fit
    fit = ProfileFitter(cp, bp, mp, pp)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    Iknown = 94.67151440319306
    Vknown = 94.67151440319306

    # Test intensity is the same
    eps = 1e-7
    assert intensity[0] == pytest.approx(Iknown, abs=eps)
    assert V[0] == pytest.approx(Vknown, abs=eps)


def test_with_flat_background_partial():
    np.random.seed(0)

    # Create profile
    p = gaussian((9, 9, 9), 1, (4, 4, 4), (2, 2, 2))
    s = flex.sum(p)
    p = p / s

    # Copy profile
    c0 = add_poisson_noise(100 * p)
    b = flex.double(flex.grid(9, 9, 9), 1)
    m = flex.bool(flex.grid(9, 9, 9), True)
    c = c0 + add_poisson_noise(b)

    # Get the partial profiles
    pp = p[0:5, :, :]
    mp = m[0:5, :, :]
    cp = c[0:5, :, :]
    bp = b[0:5, :, :]

    # Fit
    fit = ProfileFitter(cp, bp, mp, pp)
    intensity = fit.intensity()
    V = fit.variance()
    assert fit.niter() < fit.maxiter()

    Iknown = 99.06932141277105
    Vknown = 504.06932141277105

    # Test intensity is the same
    eps = 1e-7
    assert intensity[0] == pytest.approx(Iknown, abs=eps)
    assert V[0] == pytest.approx(Vknown, abs=eps)


def test_deconvolve_3_with_no_background():
    np.random.seed(0)

    I0 = [1000, 2000, 3000]

    # Create profile
    p = generate_3_profiles()

    Ical = [[], [], []]
    for it in range(1):

        # Copy profile
        c = flex.double(flex.grid(40, 9, 9))
        for i in range(p.all()[0]):
            pp = p[i : i + 1, :, :, :]
            pp.reshape(c.accessor())
            cc = add_poisson_noise(I0[i] * pp)
            c += cc
        b = flex.double(flex.grid(40, 9, 9), 0)
        m = flex.bool(flex.grid(40, 9, 9), True)

        # Fit
        fit = ProfileFitter(c, b, m, p)
        intensity = fit.intensity()
        V = fit.variance()
        assert fit.niter() < fit.maxiter()

        for i in range(3):
            Ical[i].append(intensity[i])

    for i in range(3):
        Ical[i] = sum(Ical[i]) / len(Ical[i])

    Iknown = [1048.3221116842406, 1920.9035376774107, 2938.7743506383745]

    # Test intensity is the same
    eps = 1e-7
    for i in range(3):
        assert intensity[i] == pytest.approx(Iknown[i], abs=eps)
        assert V[i] == pytest.approx(Iknown[i], abs=eps)


def test_deconvolve_3_with_flat_background():

    np.random.seed(0)

    I0 = [1000, 2000, 3000]

    # Create profile
    p = generate_3_profiles()

    Ical = [[], [], []]
    for it in range(1):

        # Copy profile
        c = flex.double(flex.grid(40, 9, 9))
        for i in range(p.all()[0]):
            pp = p[i : i + 1, :, :, :]
            pp.reshape(c.accessor())
            cc = add_poisson_noise(I0[i] * pp)
            c += cc
        b = flex.double(flex.grid(40, 9, 9), 1)
        m = flex.bool(flex.grid(40, 9, 9), True)
        c += add_poisson_noise(b)

        # Fit
        fit = ProfileFitter(c, b, m, p)
        intensity = fit.intensity()
        V = fit.variance()
        assert fit.niter() < fit.maxiter()

        for i in range(3):
            Ical[i].append(intensity[i])

    for i in range(3):
        Ical[i] = sum(Ical[i]) / len(Ical[i])

    Iknown = [1030.7018033805357, 1948.7152700952695, 2972.983204218213]
    Vknown = [4270.701803380534, 5188.715270095279, 6212.983204218214]

    # Test intensity is the same
    eps = 1e-7
    for i in range(3):
        assert intensity[i] == pytest.approx(Iknown[i], abs=eps)
        assert V[i] == pytest.approx(Vknown[i], abs=eps)


def test_deconvolve_7_with_no_background():
    np.random.seed(0)

    I0 = [1000, 1500, 2000, 2500, 3000, 3500, 4000]

    # Create profile
    p = generate_7_profiles()

    Ical = [[], [], [], [], [], [], []]
    for it in range(1):

        # Copy profile
        c = flex.double(flex.grid(40, 40, 40))
        for i in range(p.all()[0]):
            pp = p[i : i + 1, :, :, :]
            pp.reshape(c.accessor())
            cc = add_poisson_noise(I0[i] * pp)
            c += cc
        b = flex.double(flex.grid(40, 40, 40), 0)
        m = flex.bool(flex.grid(40, 40, 40), True)

        # Fit
        fit = ProfileFitter(c, b, m, p)
        intensity = fit.intensity()
        V = fit.variance()
        assert fit.niter() < fit.maxiter()

        for i in range(7):
            Ical[i].append(intensity[i])

    for i in range(7):
        Ical[i] = sum(Ical[i]) / len(Ical[i])

    Iknown = [
        1012.4193633595916,
        1472.3322059495797,
        2072.136413825826,
        2486.4902438469253,
        3012.6132119521126,
        3409.2053517072773,
        3952.803209358826,
    ]

    # Test intensity is the same
    eps = 1e-7
    for i in range(7):
        assert intensity[i] == pytest.approx(Iknown[i], abs=eps)
        assert V[i] == pytest.approx(Iknown[i], abs=eps)


def test_deconvolve_7_with_flat_background():
    np.random.seed(0)

    I0 = [1000, 1500, 2000, 2500, 3000, 3500, 4000]

    # Create profile
    p = generate_7_profiles()

    Ical = [[], [], [], [], [], [], []]
    for it in range(1):

        # Copy profile
        c = flex.double(flex.grid(40, 40, 40))
        for i in range(p.all()[0]):
            pp = p[i : i + 1, :, :, :]
            pp.reshape(c.accessor())
            cc = add_poisson_noise(I0[i] * pp)
            c += cc
        b = flex.double(flex.grid(40, 40, 40), 1)
        m = flex.bool(flex.grid(40, 40, 40), True)
        c += add_poisson_noise(b)

        # Fit
        fit = ProfileFitter(c, b, m, p)
        intensity = fit.intensity()
        V = fit.variance()
        assert fit.niter() < fit.maxiter()

        for i in range(7):
            Ical[i].append(intensity[i])

    for i in range(7):
        Ical[i] = sum(Ical[i]) / len(Ical[i])

    Iknown = [
        1042.4904898451261,
        1447.6413897568257,
        2053.2524102387893,
        2485.2789614429703,
        3063.867598583333,
        3445.9230910807582,
        3977.6744651425543,
    ]
    Vknown = [
        65042.49048984377,
        65447.6413897533,
        66053.25241023625,
        66485.27896144328,
        67063.86759858252,
        67445.92309107889,
        67977.67446514373,
    ]

    # Test intensity is the same
    eps = 1e-7
    for i in range(7):
        assert intensity[i] == pytest.approx(Iknown[i], abs=eps)
        assert V[i] == pytest.approx(Vknown[i], abs=eps)



"""Merging functions for experiment lists and reflection tables."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from io import StringIO
from typing import List, Optional

from cctbx import miller
from dxtbx.model import ExperimentList
from iotbx.merging_statistics import dataset_statistics
from mmtbx.scaling import data_statistics

from dials.algorithms.scaling.Ih_table import (
    _reflection_table_to_iobs,
    map_indices_to_asu,
)
from dials.algorithms.scaling.scaling_library import (
    merging_stats_from_scaled_array,
    scaled_data_as_miller_array,
)
from dials.algorithms.scaling.scaling_utilities import DialsMergingStatisticsError
from dials.algorithms.symmetry.absences.run_absences_checks import (
    run_systematic_absences_checks,
)
from dials.array_family import flex
from dials.report.analysis import make_merging_statistics_summary, table_1_summary
from dials.util.export_mtz import MADMergedMTZWriter, MergedMTZWriter
from dials.util.filter_reflections import filter_reflection_table

from .french_wilson import french_wilson

logger = logging.getLogger("dials")


def prepare_merged_reflection_table(
    experiments,
    reflection_table,
    d_min=None,
    d_max=None,
    partiality_threshold=0.99,
):
    """Filter the data and prepare a reflection table with merged data."""
    if (
        "inverse_scale_factor" in reflection_table
        and "intensity.scale.value" in reflection_table
    ):
        logger.info("Performing systematic absence checks on scaled data")
        reflections = filter_reflection_table(
            reflection_table,
            intensity_choice=["scale"],
            d_min=d_min,
            d_max=d_max,
            partiality_threshold=partiality_threshold,
        )
        reflections["intensity"] = reflections["intensity.scale.value"]
        reflections["variance"] = reflections["intensity.scale.variance"]
    elif "intensity.prf.value" in reflection_table:
        logger.info(
            "Performing systematic absence checks on unscaled profile-integrated data"
        )
        reflections = filter_reflection_table(
            reflection_table,
            intensity_choice=["profile"],
            d_min=d_min,
            d_max=d_max,
            partiality_threshold=partiality_threshold,
        )
        reflections["intensity"] = reflections["intensity.prf.value"]
        reflections["variance"] = reflections["intensity.prf.variance"]
    else:
        logger.info(
            "Performing systematic absence checks on unscaled summation-integrated data"
        )
        reflections = filter_reflection_table(
            reflection_table,
            intensity_choice=["sum"],
            d_min=d_min,
            d_max=d_max,
            partiality_threshold=partiality_threshold,
        )
        reflections["intensity"] = reflections["intensity.sum.value"]
        reflections["variance"] = reflections["intensity.sum.variance"]
    # now merge
    space_group = experiments[0].crystal.get_space_group()
    reflections["asu_miller_index"] = map_indices_to_asu(
        reflections["miller_index"], space_group
    )
    reflections["inverse_scale_factor"] = flex.double(reflections.size(), 1.0)
    merged = (
        _reflection_table_to_iobs(
            reflections, experiments[0].crystal.get_unit_cell(), space_group
        )
        .merge_equivalents(use_internal_variance=False)
        .array()
    )
    merged_reflections = flex.reflection_table()
    merged_reflections["intensity"] = merged.data()
    merged_reflections["variance"] = flex.pow2(merged.sigmas())
    merged_reflections["miller_index"] = merged.indices()
    return merged_reflections


class MTZDataClass:

    """Container class (i.e. Python 3.7 dataclass) for per-wavelength mtz dataset."""

    def __init__(
        self,
        wavelength=0.0,
        project_name="AUTOMATIC",
        dataset_name="NATIVE",
        crystal_name="XTAL",
        merged_array=None,
        merged_anomalous_array=None,
        amplitudes=None,
        anomalous_amplitudes=None,
        dano=None,
        multiplicities=None,
    ):
        self.wavelength = wavelength
        self.project_name = project_name
        self.dataset_name = dataset_name
        self.crystal_name = crystal_name
        self.merged_array = merged_array
        self.merged_anomalous_array = merged_anomalous_array
        self.amplitudes = amplitudes
        self.anomalous_amplitudes = anomalous_amplitudes
        self.dano = dano
        self.multiplicities = multiplicities


def make_merged_mtz_file(mtz_datasets):
    """
    Make an mtz file object for the data, adding the date, time and program.

    For multi-wavelength data, each wavelength is added as a new crystal.

    Args:
        mtz_datasets: A list of MTZDataClass objects, one per wavelength of the
            experiment.

    Returns:
        An iotbx mtz file object.
    """

    if len(mtz_datasets) > 1:
        writer = MADMergedMTZWriter
    else:
        writer = MergedMTZWriter

    mtz_writer = writer(
        mtz_datasets[0].merged_array.space_group(),
        mtz_datasets[0].merged_array.unit_cell(),
    )

    #### Add each wavelength as a new crystal.
    for dataset in mtz_datasets:
        mtz_writer.add_crystal(
            crystal_name=dataset.crystal_name, project_name=dataset.project_name
        )
        mtz_writer.add_empty_dataset(dataset.wavelength, name=dataset.dataset_name)
        mtz_writer.add_dataset(
            dataset.merged_array,
            dataset.merged_anomalous_array,
            dataset.amplitudes,
            dataset.anomalous_amplitudes,
            dataset.dano,
            dataset.multiplicities,
        )

    return mtz_writer.mtz_file


@dataclass
class MergingStatisticsData:
    reflections: List[flex.reflection_table]
    experiments: ExperimentList
    scaled_miller_array: miller.array
    merging_statistics_result: Optional[dataset_statistics] = None
    anom_merging_statistics_result: Optional[dataset_statistics] = None
    anomalous_amplitudes: Optional[miller.array] = None
    Wilson_B_iso: Optional[float] = None

    def __str__(self):
        if not self.merging_statistics_result:
            return ""
        stats_summary = make_merging_statistics_summary(self.merging_statistics_result)
        stats_summary += table_1_summary(
            self.merging_statistics_result,
            self.anom_merging_statistics_result,
            Wilson_B_iso=self.Wilson_B_iso,
        )
        return stats_summary

    def __repr__(self):
        return self.__str__()


def merge(
    experiments,
    reflections,
    d_min=None,
    d_max=None,
    combine_partials=True,
    partiality_threshold=0.4,
    best_unit_cell=None,
    anomalous=True,
    use_internal_variance=False,
    assess_space_group=False,
    n_bins=20,
):
    """
    Merge reflection table data and generate a summary of the merging statistics.

    This procedure filters the input data, merges the data (normal and optionally
    anomalous), assesses the space group symmetry and generates a summary
    of the merging statistics.

    Returns two merge_equivalents objects and a statistics summary.
    """

    logger.info("\nMerging scaled reflection data\n")
    # first filter bad reflections using dials.util.filter methods
    reflections = filter_reflection_table(
        reflections,
        intensity_choice=["scale"],
        d_min=d_min,
        d_max=d_max,
        combine_partials=combine_partials,
        partiality_threshold=partiality_threshold,
    )
    # ^ scale factor has been applied, so now set to 1.0 - okay as not
    # going to output scale factor in merged mtz.
    reflections["inverse_scale_factor"] = flex.double(reflections.size(), 1.0)
    scaled_array = scaled_data_as_miller_array(
        [reflections], experiments, best_unit_cell
    )
    # Note, merge_equivalents does not raise an error if data is unique.
    merged = scaled_array.merge_equivalents(use_internal_variance=use_internal_variance)
    merged_anom = None

    if anomalous:
        anomalous_scaled = scaled_array.as_anomalous_array()
        merged_anom = anomalous_scaled.merge_equivalents(
            use_internal_variance=use_internal_variance
        )

    # Before merge, do assessment of the space_group
    if assess_space_group:
        merged_reflections = flex.reflection_table()
        merged_reflections["intensity"] = merged.array().data()
        merged_reflections["variance"] = flex.pow2(merged.array().sigmas())
        merged_reflections["miller_index"] = merged.array().indices()
        logger.info("Running systematic absences check")
        run_systematic_absences_checks(experiments, merged_reflections)

    stats_data = MergingStatisticsData([reflections], experiments, scaled_array)

    try:
        stats, anom_stats = merging_stats_from_scaled_array(
            scaled_array,
            n_bins,
            use_internal_variance,
        )
    except DialsMergingStatisticsError as e:
        logger.error(e, exc_info=True)
    else:
        stats_data.merging_statistics_result = stats
        stats_data.anom_merging_statistics_result = anom_stats

    return merged, merged_anom, stats_data


def show_wilson_scaling_analysis(merged_intensities, n_residues=200):
    """
    Report the wilson statistics for a merged intensity array

    Args:
        merged_intensities: A merged miller intensity array.
        n_residues: The number of residues to use for the wilson analysis.
    """
    if not merged_intensities.space_group().is_centric():
        try:
            wilson_scaling = data_statistics.wilson_scaling(
                miller_array=merged_intensities, n_residues=n_residues
            )
        except (IndexError, RuntimeError) as e:
            logger.error(
                "\n"
                "Error encountered during Wilson statistics calculation:\n"
                "Perhaps there are too few unique reflections.\n"
                "%s",
                e,
                exc_info=True,
            )
        else:
            # Divert output through logger - do with StringIO rather than
            # info_handle else get way too much whitespace in output.
            out = StringIO()
            wilson_scaling.show(out=out)
            logger.info(out.getvalue())
            return wilson_scaling.iso_b_wilson
    return None


def truncate(
    merged_intensities,
    implementation: str = "dials",
    min_reflections: int = 200,
    fallback_to_flat_prior: bool = True,
):
    """
    Perform French-Wilson truncation procedure on merged intensities.

    Args:
        merged_intensities (miller.array): A merged miller intensity array (normal or anomalous)
        implementation (str): Choice of implementation of French & Wilson algorithm, either
                              "dials" or "cctbx"
        min_reflections (int): Minimum number of reflections to perform the French & Wilson
                               procedure
        fallback_to_flat_prior (bool): Fallback to assumption of a flat, positive prior,
                                       if the number of reflections are fewer than min_reflections,
                                       i.e. |F| = sqrt((Io+sqrt(Io**2 +2sigma**2))/2.0)

    Returns:
        (tuple): tuple containing:
            amplitudes: A normal all-positive miller amplitude array
            anom_amplitudes: An anomalous all-positive amplitude array, if the
                input array has the anomalous_flag set, else None.
            dano: The array of anomalous differences, if the input array has the
                anomalous_flag set, else None.
    """
    logger.info("\nPerforming French-Wilson treatment of scaled intensities")
    out = StringIO()
    n_refl = merged_intensities.size()
    if n_refl < min_reflections and fallback_to_flat_prior:
        logger.info(
            "Insufficient reflections for French & Wilson procedure, "
            "falling back to assumption of a flat, positive prior, i.e.: "
            "  |F| = sqrt((Io+sqrt(Io**2 +2sigma**2))/2.0)"
        )
        do_french_wilson = lambda ma: ma.enforce_positive_amplitudes()
    elif n_refl < min_reflections:
        raise ValueError(
            "Insufficient reflections for French & Wilson procedure. "
            "Either set fallback_to_flat_prior=True or truncate=False."
        )
    elif implementation == "cctbx":
        do_french_wilson = lambda ma: ma.french_wilson(log=out)
    else:
        do_french_wilson = french_wilson

    if merged_intensities.anomalous_flag():
        anom_amplitudes = do_french_wilson(merged_intensities)
        n_removed = merged_intensities.size() - anom_amplitudes.size()
        assert anom_amplitudes.is_xray_amplitude_array()
        amplitudes = anom_amplitudes.as_non_anomalous_array()
        amplitudes = amplitudes.merge_equivalents(use_internal_variance=False).array()
        dano = anom_amplitudes.anomalous_differences()
    else:
        anom_amplitudes = None
        dano = None
        amplitudes = do_french_wilson(merged_intensities)
        n_removed = merged_intensities.size() - amplitudes.size()
    logger.info("Total number of rejected intensities %s", n_removed)
    logger.debug(out.getvalue())
    return amplitudes, anom_amplitudes, dano



from __future__ import annotations


def filter_shadowed_reflections(experiments, reflections, experiment_goniometer=False):
    from dxtbx.masking import is_inside_polygon
    from scitbx.array_family import flex

    shadowed = flex.bool(reflections.size(), False)
    for expt_id in range(len(experiments)):
        expt = experiments[expt_id]
        imageset = expt.imageset
        masker = imageset.masker()
        detector = expt.detector
        sel = reflections["id"] == expt_id
        isel = sel.iselection()
        x, y, z = reflections["xyzcal.px"].select(isel).parts()
        start, end = expt.scan.get_array_range()
        for i in range(start, end):
            shadow = masker.project_extrema(
                detector, expt.scan.get_angle_from_array_index(i)
            )
            img_sel = (z >= i) & (z < (i + 1))
            img_isel = img_sel.iselection()
            for p_id in range(len(detector)):
                panel = reflections["panel"].select(img_isel)
                if shadow[p_id].size() < 4:
                    continue
                panel_isel = img_isel.select(panel == p_id)
                inside = is_inside_polygon(
                    shadow[p_id],
                    flex.vec2_double(
                        x.select(isel.select(panel_isel)),
                        y.select(isel.select(panel_isel)),
                    ),
                )
                shadowed.set_selected(panel_isel, inside)

    return shadowed



from __future__ import annotations

import os
import pickle

import iotbx.phil
from dxtbx.imageset import ImageSet
from libtbx.math_utils import nearest_integer as nint
from scitbx.array_family import flex

from dials.algorithms.image.threshold import DispersionThresholdDebug
from dials.util import Sorry, show_mail_handle_errors
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """

This program can be used to estimate the gain of the detector. For pixel array
detectors the gain is usually set to 1.00. This means that the pixels behave
according to Poisson statistics. However, for older CCD detectors the gain may
have a different value. This value is important because it can affect, amongst
other things, the ability of the spot finding algorithm which can result in
noise being identified as diffraction spots.

Examples::

  dials.estimate_gain models.expt
"""

phil_scope = iotbx.phil.parse(
    """\
  kernel_size = 10,10
    .type = ints(size=2, value_min=1)
  max_images = 1
    .type = int
    .help = "For multi-file images (NeXus for example), report a gain for each"
            "image, up to max_images, and then report an average gain"
  output {
    gain_map = None
      .type = str
      .help = "Name of output gain map file"
  }
""",
    process_includes=True,
)


def estimate_gain(
    imageset: ImageSet,
    kernel_size: tuple[int, int] = (10, 10),
    output_gain_map: str | bytes | os.PathLike | None = None,
    max_images: int = 1,
) -> float:
    detector = imageset.get_detector()

    gains = flex.double()

    for image_no in range(len(imageset)):
        raw_data = imageset.get_raw_data(image_no)

        gain_value = 1
        gain_map = [
            flex.double(raw_data[i].accessor(), gain_value)
            for i in range(len(detector))
        ]

        mask = imageset.get_mask(image_no)

        min_local = 0

        # dummy values, shouldn't affect results
        nsigma_b = 6
        nsigma_s = 3
        global_threshold = 0

        kabsch_debug_list = [
            DispersionThresholdDebug(
                raw_data[i_panel].as_double(),
                mask[i_panel],
                gain_map[i_panel],
                kernel_size,
                nsigma_b,
                nsigma_s,
                global_threshold,
                min_local,
            )
            for i_panel in range(len(detector))
        ]

        dispersion = flex.double()
        for kabsch in kabsch_debug_list:
            dispersion.extend(kabsch.index_of_dispersion().as_1d())

        sorted_dispersion = flex.sorted(dispersion)

        q1 = sorted_dispersion[nint(len(sorted_dispersion) / 4)]
        q2 = sorted_dispersion[nint(len(sorted_dispersion) / 2)]
        q3 = sorted_dispersion[nint(len(sorted_dispersion) * 3 / 4)]
        iqr = q3 - q1

        print(f"q1, q2, q3: {q1:.2f}, {q2:.2f}, {q3:.2f}")
        if iqr == 0.0:
            raise Sorry("Unable to robustly estimate the variation of pixel values.")

        inlier_sel = (sorted_dispersion > (q1 - 1.5 * iqr)) & (
            sorted_dispersion < (q3 + 1.5 * iqr)
        )
        sorted_dispersion = sorted_dispersion.select(inlier_sel)
        gain = sorted_dispersion[nint(len(sorted_dispersion) / 2)]
        print(f"Estimated gain: {gain:.2f}")
        gains.append(gain)

        if image_no == 0:
            gain0 = gain
        if image_no + 1 >= max_images:
            break

    if len(gains) > 1:
        stats = flex.mean_and_variance(gains)
        print(
            "Average gain: %.2f +/- %.2f"
            % (stats.mean(), stats.unweighted_sample_standard_deviation())
        )

    if output_gain_map:
        if len(gains) > 1:
            raw_data = imageset.get_raw_data(0)
        # write the gain map
        gain_map_flex = flex.double(flex.grid(raw_data[0].all()), gain0)
        with open(output_gain_map, "wb") as fh:
            pickle.dump(gain_map_flex, fh, protocol=pickle.HIGHEST_PROTOCOL)

    return gain0


@show_mail_handle_errors()
def run(args: list[str] | None = None) -> None:
    usage = "dials.estimate_gain [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=True,
        read_experiments_from_images=True,
        epilog=help_message,
    )

    params, _ = parser.parse_args(args, show_diff_phil=False)

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        print("The following parameters have been modified:\n")
        print(diff_phil)

    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        return
    elif len(experiments.imagesets()) > 1:
        raise Sorry("Only one imageset can be processed at a time")
    else:
        imagesets = experiments.imagesets()

    assert len(imagesets) == 1
    imageset = imagesets[0]
    estimate_gain(
        imageset, params.kernel_size, params.output.gain_map, params.max_images
    )


if __name__ == "__main__":
    run()



from __future__ import annotations

import math
from pathlib import Path

import pytest

from scitbx import matrix


def test(dials_regression, tmp_path):
    import dxtbx
    from iotbx.xds import integrate_hkl, xparm
    from rstbx.cftbx.coordinate_frame_converter import coordinate_frame_converter

    from dials.algorithms.spot_prediction import RotationAngles

    # The XDS files to read from
    integrate_filename = Path(dials_regression) / "data" / "sim_mx" / "INTEGRATE.HKL"
    gxparm_filename = Path(dials_regression) / "data" / "sim_mx" / "GXPARM.XDS"

    # Read the XDS files
    integrate_handle = integrate_hkl.reader()
    integrate_handle.read_file(integrate_filename)
    gxparm_handle = xparm.reader()
    gxparm_handle.read_file(gxparm_filename)

    # Get the parameters we need from the GXPARM file
    models = dxtbx.load(gxparm_filename)
    beam = models.get_beam()
    gonio = models.get_goniometer()
    scan = models.get_scan()

    # Get the crystal parameters
    cfc = coordinate_frame_converter(gxparm_filename)
    a_vec = cfc.get("real_space_a")
    b_vec = cfc.get("real_space_b")
    c_vec = cfc.get("real_space_c")
    UB = matrix.sqr(a_vec + b_vec + c_vec).inverse()
    ub_matrix = UB

    # Get the number of frames from the max z value
    xcal, ycal, zcal = zip(*integrate_handle.xyzcal)
    num_frames = int(math.ceil(max(zcal)))
    scan.set_image_range(
        (scan.get_image_range()[0], scan.get_image_range()[0] + num_frames - 1)
    )

    # Create the rotation angle object
    ra = RotationAngles(beam.get_s0(), gonio.get_rotation_axis())

    # Setup the matrices
    ub = matrix.sqr(ub_matrix)
    s0 = matrix.col(beam.get_s0())
    m2 = matrix.col(gonio.get_rotation_axis())

    # For all the miller indices
    for h in integrate_handle.hkl:
        h = matrix.col(h)

        # Calculate the angles
        angles = ra(h, ub)

        # For all the angles
        for phi in angles:
            r = m2.axis_and_angle_as_r3_rotation_matrix(angle=phi)
            pstar = r * ub * h
            s1 = s0 + pstar
            assert s1.length() == pytest.approx(s0.length(), abs=1e-7)

    # Create a dict of lists of xy for each hkl
    gen_phi = {}
    for h in integrate_handle.hkl:

        # Calculate the angles
        angles = ra(h, ub)
        gen_phi[h] = angles
    #        for phi in angles:
    #            try:
    #                a = gen_phi[h]
    #                a.append(phi)
    #                gen_phi[h] = a
    #            except KeyError:
    #                gen_phi[h] = [phi]

    # For each hkl in the xds file
    for hkl, xyz in zip(integrate_handle.hkl, integrate_handle.xyzcal):

        # Calculate the XDS phi value
        xds_phi = (
            scan.get_oscillation(deg=False)[0]
            + xyz[2] * scan.get_oscillation(deg=False)[1]
        )

        # Select the nearest xy to use if there are 2
        my_phi = gen_phi[hkl]
        if len(my_phi) == 2:
            my_phi0 = my_phi[0]
            my_phi1 = my_phi[1]
            diff0 = abs(xds_phi - my_phi0)
            diff1 = abs(xds_phi - my_phi1)
            if diff0 < diff1:
                my_phi = my_phi0
            else:
                my_phi = my_phi1
        else:
            my_phi = my_phi[0]

        # Check the Phi values are the same
        assert xds_phi == pytest.approx(my_phi, abs=0.1)



# LIBTBX_SET_DISPATCHER_NAME dials.import

from __future__ import annotations

import logging
import pickle
import sys
from collections import namedtuple

import dxtbx.model.compare as compare
import libtbx.phil
from dxtbx.imageset import ImageGrid, ImageSequence
from dxtbx.model.experiment_list import (
    Experiment,
    ExperimentList,
    ExperimentListFactory,
)

from dials.util import Sorry, log, show_mail_handle_errors
from dials.util.multi_dataset_handling import generate_experiment_identifiers
from dials.util.options import ArgumentParser, flatten_experiments
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.import")


def _pickle_load(fh):
    return pickle.load(fh, encoding="bytes")


help_message = """

This program is used to import image data files into a format that can be used
within dials. The program looks at the metadata for each image along with the
filenames to determine the relationship between sets of images. Once all the
images have been analysed, a experiments object is written to file which specifies
the relationship between files. For example if two sets of images which belong
to two rotation scans have been given, two image sequences will be saved. Images to
be processed are specified as command line arguments. Sometimes, there is a
maximum number of arguments that can be given on the command line and the number
of files may exceed this. In this case image filenames can be input on stdin
as shown in the examples below. Alternatively a template can be specified using
the template= parameter where the consecutive digits representing the image
numbers in the filenames are replaced with '#' characters.

The geometry can be set manually, either by using the reference_geometry=
parameter to specify an experiment list .expt file containing
the reference geometry, by using the mosflm_beam_centre= parameter to set
the Mosflm beam centre, or by specifying each variable to be overridden
using various geometry parameters.

Examples::

  dials.import /data/directory-containing-images/

  dials.import image_*.cbf

  dials.import image_1_*.cbf image_2_*.cbf

  dials.import directory/with/images

  dials.import template=image_1_####.cbf

  dials.import directory=directory/with/images

  find . -name "image_*.cbf" | dials.import

  dials.import << EOF
  image_1.cbf
  image_2.cbf
  EOF
"""


# Create the phil parameters
phil_scope = libtbx.phil.parse(
    """

  output {

    experiments = imported.expt
      .type = str
      .help = "The output experiment file"

    log = 'dials.import.log'
      .type = str
      .help = "The log filename"

    compact = False
      .type = bool
      .help = "For experiment output use compact JSON representation"

  }

  identifier_type = *uuid timestamp None
    .type = choice
    .help = "Type of unique identifier to generate."

  input {

    ignore_unhandled = True
      .type = bool
      .help = "Ignore unhandled input (e.g. log files)"

    template = None
      .type = str
      .help = "The image sequence template"
      .multiple = True

    directory = None
      .type = str
      .help = "A directory with images"
      .multiple = True

    reference_geometry = None
      .type = path
      .help = "Experimental geometry from this models.expt "
              "will override the geometry from the "
              "image headers."

    check_reference_geometry = True
      .type = bool
      .expert_level = 2
      .help = "If True, assert the reference geometry is similar to"
              "the image geometry"

    use_beam_reference = True
      .type = bool
      .expert_level = 2
      .help = "If True, the beam from reference_geometry will override "
              "the beam from the image headers."

    use_gonio_reference = True
      .type = bool
      .expert_level = 2
      .help = "If True, the goniometer from reference_geometry will override "
              "the goniometer from the image headers."

    use_detector_reference = True
      .type = bool
      .expert_level = 2
      .help = "If True, the detector from reference_geometry will override "
              "the detector from the image headers."

    allow_multiple_sequences = True
      .type = bool
      .help = "If False, raise an error if multiple sequences are found"

    as_grid_scan = False
      .type = bool
      .help = "Import as grid scan"

    grid_size = None
      .type = ints(size=2)
      .help = "If importing as a grid scan set the size"
  }

  include scope dials.util.options.format_phil_scope

  include scope dials.util.options.geometry_phil_scope

  lookup {
    mask = None
      .type = str
      .help = "Apply a mask to the imported data"

    gain = None
      .type = str
      .help = "Apply a gain to the imported data"

    pedestal = None
      .type = str
      .help = "Apply a pedestal to the imported data"

    dx = None
      .type = str
      .help = "Apply an x geometry offset"
              "If both dx and dy are set then"
              "OffsetParallaxCorrectedPxMmStrategy will be used"

    dy = None
      .type = str
      .help = "Apply an y geometry offset"
              "If both dx and dy are set then"
              "OffsetParallaxCorrectedPxMmStrategy will be used"
  }
""",
    process_includes=True,
)


def _extract_or_read_imagesets(params):
    """
    Return a list of ImageSets, importing them via alternative means if necessary.

    The "Alternative Means" means via params.input.template or .directory,
    if the images to import haven't been specified directly.

    Args:
        params: The phil.scope_extract from dials.import

    Returns: A list of ImageSet objects
    """

    # Get the experiments
    experiments = flatten_experiments(params.input.experiments)

    # Check we have some filenames
    if len(experiments) == 0:

        # FIXME Should probably make this smarter since it requires editing here
        # and in dials.import phil scope
        try:
            format_kwargs = {
                "dynamic_shadowing": params.format.dynamic_shadowing,
                "multi_panel": params.format.multi_panel,
            }
        except AttributeError:
            format_kwargs = None

        # Check if a template has been set and print help if not, otherwise try to
        # import the images based on the template input
        if len(params.input.template) > 0:

            experiments = ExperimentListFactory.from_templates(
                params.input.template,
                image_range=params.geometry.scan.image_range,
                format_kwargs=format_kwargs,
            )
            if len(experiments) == 0:
                raise Sorry(
                    "No experiments found matching template %s"
                    % params.input.experiments
                )
        elif len(params.input.directory) > 0:
            experiments = ExperimentListFactory.from_filenames(
                params.input.directory, format_kwargs=format_kwargs
            )
            if len(experiments) == 0:
                raise Sorry(
                    "No experiments found in directories %s" % params.input.directory
                )
        else:
            raise Sorry("No experiments found")

    # TODO (Nick):  This looks redundant as the experiments are immediately discarded.
    #               verify this, and remove if it is.
    if params.identifier_type:
        generate_experiment_identifiers(experiments, params.identifier_type)

    # Get a list of all imagesets
    imageset_list = experiments.imagesets()

    # Return the experiments
    return imageset_list


class ReferenceGeometryUpdater:
    """
    A class to replace beam + detector with a reference
    """

    def __init__(self, params):
        """
        Load the reference geometry
        """
        self.params = params
        self.reference = self.load_reference_geometry(params)

    def __call__(self, imageset):
        """
        Replace with the reference geometry
        """
        if self.params.input.check_reference_geometry:
            # Check static detector items are the same
            assert self.reference.detector.is_similar_to(
                imageset.get_detector(), static_only=True
            ), "Reference detector model does not match input detector model"

        # Set beam and detector
        if self.params.input.use_beam_reference:
            imageset.set_beam(self.reference.beam)
        if self.params.input.use_detector_reference:
            imageset.set_detector(self.reference.detector)
        if self.params.input.use_gonio_reference:
            imageset.set_goniometer(self.reference.goniometer)
        return imageset

    def load_reference_geometry(self, params):
        """
        Load a reference geometry file
        """
        # Load reference geometry
        reference_detector = None
        reference_beam = None
        if params.input.reference_geometry is not None:
            from dxtbx.serialize import load

            experiments = None
            experiments = load.experiment_list(
                params.input.reference_geometry, check_format=False
            )
            assert experiments, "Could not import reference geometry"
            assert len(experiments.detectors()) >= 1
            assert len(experiments.beams()) >= 1
            if len(experiments.detectors()) > 1:
                raise Sorry(
                    "The reference geometry file contains %d detector definitions, but only a single definition is allowed."
                    % len(experiments.detectors())
                )
            if len(experiments.beams()) > 1:
                raise Sorry(
                    "The reference geometry file contains %d beam definitions, but only a single definition is allowed."
                    % len(experiments.beams())
                )
            reference_detector = experiments.detectors()[0]
            reference_beam = experiments.beams()[0]
            reference_goniometer = experiments.goniometers()[0]
        Reference = namedtuple("Reference", ["detector", "beam", "goniometer"])
        return Reference(
            detector=reference_detector,
            beam=reference_beam,
            goniometer=reference_goniometer,
        )


class ManualGeometryUpdater:
    """
    A class to update the geometry manually
    """

    def __init__(self, params):
        """
        Save the params
        """
        self.params = params

    def __call__(self, imageset):
        """
        Override the parameters
        """
        from copy import deepcopy

        from dxtbx.imageset import ImageSequence, ImageSetFactory
        from dxtbx.model import (
            BeamFactory,
            DetectorFactory,
            GoniometerFactory,
            ScanFactory,
        )

        if self.params.geometry.convert_sequences_to_stills:
            imageset = ImageSetFactory.imageset_from_anyset(imageset)
            for j in range(len(imageset)):
                imageset.set_scan(None, j)
                imageset.set_goniometer(None, j)
        if not isinstance(imageset, ImageSequence):
            if self.params.geometry.convert_stills_to_sequences:
                imageset = self.convert_stills_to_sequence(imageset)
        if isinstance(imageset, ImageSequence):
            beam = BeamFactory.from_phil(self.params.geometry, imageset.get_beam())
            detector = DetectorFactory.from_phil(
                self.params.geometry, imageset.get_detector(), beam
            )
            goniometer = GoniometerFactory.from_phil(
                self.params.geometry, imageset.get_goniometer()
            )
            scan = ScanFactory.from_phil(
                self.params.geometry, deepcopy(imageset.get_scan())
            )
            i0, i1 = scan.get_array_range()
            j0, j1 = imageset.get_scan().get_array_range()
            if i0 < j0 or i1 > j1:
                imageset = self.extrapolate_imageset(
                    imageset=imageset,
                    beam=beam,
                    detector=detector,
                    goniometer=goniometer,
                    scan=scan,
                )
            else:
                imageset.set_beam(beam)
                imageset.set_detector(detector)
                imageset.set_goniometer(goniometer)
                imageset.set_scan(scan)
        else:
            for i in range(len(imageset)):
                beam = BeamFactory.from_phil(self.params.geometry, imageset.get_beam(i))
                detector = DetectorFactory.from_phil(
                    self.params.geometry, imageset.get_detector(i), beam
                )
                goniometer = GoniometerFactory.from_phil(
                    self.params.geometry, imageset.get_goniometer(i)
                )
                scan = ScanFactory.from_phil(self.params.geometry, imageset.get_scan(i))
                imageset.set_beam(beam, i)
                imageset.set_detector(detector, i)
                imageset.set_goniometer(goniometer, i)
                imageset.set_scan(scan, i)
        return imageset

    def extrapolate_imageset(
        self, imageset=None, beam=None, detector=None, goniometer=None, scan=None
    ):
        from dxtbx.imageset import ImageSetFactory

        first, last = scan.get_image_range()
        sequence = ImageSetFactory.make_sequence(
            template=imageset.get_template(),
            indices=list(range(first, last + 1)),
            format_class=imageset.get_format_class(),
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            format_kwargs=imageset.params(),
        )
        return sequence

    def convert_stills_to_sequence(self, imageset):
        from dxtbx.model import Scan

        assert self.params.geometry.scan.oscillation is not None
        beam = imageset.get_beam(index=0)
        detector = imageset.get_detector(index=0)
        goniometer = imageset.get_goniometer(index=0)
        for i in range(1, len(imageset)):
            b_i = imageset.get_beam(i)
            d_i = imageset.get_detector(i)
            g_i = imageset.get_goniometer(i)
            assert (beam is None and b_i is None) or beam.is_similar_to(
                imageset.get_beam(index=i),
                wavelength_tolerance=self.params.input.tolerance.beam.wavelength,
                direction_tolerance=self.params.input.tolerance.beam.direction,
                polarization_normal_tolerance=self.params.input.tolerance.beam.polarization_normal,
                polarization_fraction_tolerance=self.params.input.tolerance.beam.polarization_fraction,
            )
            assert (detector is None and d_i is None) or detector.is_similar_to(
                imageset.get_detector(index=i),
                fast_axis_tolerance=self.params.input.tolerance.detector.fast_axis,
                slow_axis_tolerance=self.params.input.tolerance.detector.slow_axis,
                origin_tolerance=self.params.input.tolerance.detector.origin,
            )
            assert (goniometer is None and g_i is None) or goniometer.is_similar_to(
                imageset.get_goniometer(index=i),
                rotation_axis_tolerance=self.params.input.tolerance.goniometer.rotation_axis,
                fixed_rotation_tolerance=self.params.input.tolerance.goniometer.fixed_rotation,
                setting_rotation_tolerance=self.params.input.tolerance.goniometer.setting_rotation,
            )
        oscillation = self.params.geometry.scan.oscillation
        from dxtbx.imageset import ImageSetFactory
        from dxtbx.sequence_filenames import template_regex_from_list

        template, indices = template_regex_from_list(imageset.paths())
        image_range = (min(indices), max(indices))
        assert (image_range[1] + 1 - image_range[0]) == len(indices)
        scan = Scan(image_range=image_range, oscillation=oscillation)
        if template is None:
            paths = [imageset.get_path(i) for i in range(len(imageset))]
            assert len(set(paths)) == 1
            template = paths[0]
        new_sequence = ImageSetFactory.make_sequence(
            template=template,
            indices=indices,
            format_class=imageset.reader().get_format_class(),
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
        )
        return new_sequence


class MetaDataUpdater:
    """
    A class to manage updating the experiments metadata
    """

    def __init__(self, params):
        """
        Init the class
        """
        from dials.util.options import geometry_phil_scope

        self.params = params

        # Create the geometry updater
        self.update_geometry = []
        update_order = []

        # First add reference geometry is present
        if self.params.input.reference_geometry is not None:
            self.update_geometry.append(ReferenceGeometryUpdater(self.params))
            update_order.append("Reference geometry")

        # Then add manual geometry
        working_phil = geometry_phil_scope.format(self.params)
        diff_phil = geometry_phil_scope.fetch_diff(source=working_phil)
        if diff_phil.as_str() != "":
            self.update_geometry.append(ManualGeometryUpdater(self.params))
            update_order.append("Manual geometry")

        if len(update_order) > 0:
            logger.info("")
            logger.info("Applying input geometry in the following order:")
            for i, item in enumerate(update_order, start=1):
                logger.info("  %d. %s", i, item)
            logger.info("")

    def __call__(self, imageset_list):
        """
        Transform the metadata
        """
        # Import the lookup data
        lookup = self.import_lookup_data(self.params)

        # Convert all to ImageGrid
        if self.params.input.as_grid_scan:
            imageset_list = self.convert_to_grid_scan(imageset_list, self.params)

        # Create the experiments
        experiments = ExperimentList()

        # Loop through imagesets
        for imageset in imageset_list:

            # Set the external lookups
            imageset = self.update_lookup(imageset, lookup)

            # Update the geometry
            for updater in self.update_geometry:
                imageset = updater(imageset)

            # Check beam and detector are present
            if imageset.get_beam() is None or imageset.get_detector() is None:
                raise Sorry(
                    """
          Imageset contains no beam or detector model. This means you will be
          unable to process your data.

          Possible causes of this error are:
             - A problem reading the images with one of the dxtbx format classes
             - A lack of header information in the file itself.

          You can override this by specifying the metadata as geometry parameters
        """
                )

            # Check if dx and dy are set
            if [
                imageset.external_lookup.dx.filename,
                imageset.external_lookup.dy.filename,
            ].count(None) == 0:
                imageset.update_detector_px_mm_data()
            elif [
                imageset.external_lookup.dx.filename,
                imageset.external_lookup.dy.filename,
            ].count(None) == 1:
                raise Sorry(
                    """
          Only 1 offset map is set. Need to set both dx and d
        """
                )

            # Append to new imageset list
            if isinstance(imageset, ImageSequence):
                if imageset.get_scan().is_still():
                    # make lots of experiments all pointing at one
                    # image set

                    # check if user has overridden the input - if yes, recall
                    # that these are in people numbers (1...) and are inclusive
                    if self.params.geometry.scan.image_range:
                        user_start, user_end = self.params.geometry.scan.image_range
                        offset = imageset.get_scan().get_array_range()[0]
                        start, end = user_start - 1, user_end
                    else:
                        start, end = imageset.get_scan().get_array_range()
                        offset = 0

                    for j in range(start, end):
                        subset = imageset[j - offset : j - offset + 1]
                        experiments.append(
                            Experiment(
                                imageset=imageset,
                                beam=imageset.get_beam(),
                                detector=imageset.get_detector(),
                                goniometer=imageset.get_goniometer(),
                                scan=subset.get_scan(),
                                crystal=None,
                            )
                        )
                else:
                    # have just one experiment
                    experiments.append(
                        Experiment(
                            imageset=imageset,
                            beam=imageset.get_beam(),
                            detector=imageset.get_detector(),
                            goniometer=imageset.get_goniometer(),
                            scan=imageset.get_scan(),
                            crystal=None,
                        )
                    )
            else:
                for i in range(len(imageset)):
                    experiments.append(
                        Experiment(
                            imageset=imageset[i : i + 1],
                            beam=imageset.get_beam(i),
                            detector=imageset.get_detector(i),
                            goniometer=imageset.get_goniometer(i),
                            scan=imageset.get_scan(i),
                            crystal=None,
                        )
                    )

        if self.params.identifier_type:
            generate_experiment_identifiers(experiments, self.params.identifier_type)
        # Return the experiments
        return experiments

    def update_lookup(self, imageset, lookup):
        from dxtbx.format.image import ImageBool, ImageDouble

        if lookup.size is not None:
            d = imageset.get_detector()
            assert len(lookup.size) == len(d), "Incompatible size"
            for s, p in zip(lookup.size, d):
                assert s == p.get_image_size()[::-1], "Incompatible size"
            if lookup.mask.filename is not None:
                imageset.external_lookup.mask.filename = lookup.mask.filename
                imageset.external_lookup.mask.data = ImageBool(lookup.mask.data)
            if lookup.gain.filename is not None:
                imageset.external_lookup.gain.filename = lookup.gain.filename
                imageset.external_lookup.gain.data = ImageDouble(lookup.gain.data)
            if lookup.dark.filename is not None:
                imageset.external_lookup.pedestal.filename = lookup.dark.filename
                imageset.external_lookup.pedestal.data = ImageDouble(lookup.dark.data)
            if lookup.dx.filename is not None:
                imageset.external_lookup.dx.filename = lookup.dx.filename
                imageset.external_lookup.dx.data = ImageDouble(lookup.dx.data)
            if lookup.dy.filename is not None:
                imageset.external_lookup.dy.filename = lookup.dy.filename
                imageset.external_lookup.dy.data = ImageDouble(lookup.dy.data)
        return imageset

    def import_lookup_data(self, params):
        """
        Get the lookup data
        """
        # Check the lookup inputs
        mask_filename = None
        gain_filename = None
        dark_filename = None
        dx_filename = None
        dy_filename = None
        mask = None
        gain = None
        dark = None
        dx = None
        dy = None
        lookup_size = None
        if params.lookup.mask is not None:
            mask_filename = params.lookup.mask
            with open(mask_filename, "rb") as fh:
                mask = _pickle_load(fh)
            if not isinstance(mask, tuple):
                mask = (mask,)
            lookup_size = [m.all() for m in mask]
        if params.lookup.gain is not None:
            gain_filename = params.lookup.gain
            with open(gain_filename, "rb") as fh:
                gain = _pickle_load(fh)
            if not isinstance(gain, tuple):
                gain = (gain,)
            if lookup_size is None:
                lookup_size = [g.all() for g in gain]
            else:
                assert len(gain) == len(lookup_size), "Incompatible size"
                for s, g in zip(lookup_size, gain):
                    assert s == g.all(), "Incompatible size"
        if params.lookup.pedestal is not None:
            dark_filename = params.lookup.pedestal
            with open(dark_filename, "rb") as fh:
                dark = _pickle_load(fh)
            if not isinstance(dark, tuple):
                dark = (dark,)
            if lookup_size is None:
                lookup_size = [d.all() for d in dark]
            else:
                assert len(dark) == len(lookup_size), "Incompatible size"
                for s, d in zip(lookup_size, dark):
                    assert s == d.all(), "Incompatible size"
        if params.lookup.dx is not None:
            dx_filename = params.lookup.dx
            with open(dx_filename, "rb") as fh:
                dx = _pickle_load(fh)
            if not isinstance(dx, tuple):
                dx = (dx,)
            if lookup_size is None:
                lookup_size = [d.all() for d in dx]
            else:
                assert len(dx) == len(lookup_size), "Incompatible size"
                for s, d in zip(lookup_size, dx):
                    assert s == d.all(), "Incompatible size"
        if params.lookup.dy is not None:
            dy_filename = params.lookup.dy
            with open(dx_filename, "rb") as fh:
                dy = _pickle_load(fh)
            if not isinstance(dy, tuple):
                dy = (dy,)
            if lookup_size is None:
                lookup_size = [d.all() for d in dy]
            else:
                assert len(dy) == len(lookup_size), "Incompatible size"
                for s, d in zip(lookup_size, dy):
                    assert s == d.all(), "Incompatible size"
        Lookup = namedtuple("Lookup", ["size", "mask", "gain", "dark", "dx", "dy"])
        Item = namedtuple("Item", ["data", "filename"])
        return Lookup(
            size=lookup_size,
            mask=Item(data=mask, filename=mask_filename),
            gain=Item(data=gain, filename=gain_filename),
            dark=Item(data=dark, filename=dark_filename),
            dx=Item(data=dx, filename=dx_filename),
            dy=Item(data=dy, filename=dy_filename),
        )

    def convert_to_grid_scan(self, imageset_list, params):
        """
        Convert the imagesets to grid scans
        """
        if params.input.grid_size is None:
            raise Sorry("The input.grid_size parameter is required")
        result = []
        for imageset in imageset_list:
            result.append(
                ImageGrid.from_imageset(imageset.as_imageset(), params.input.grid_size)
            )
        return result


def print_sequence_diff(sequence1, sequence2, params):
    """
    Print a diff between sequences.
    """
    logger.info(
        compare.sequence_diff(sequence1, sequence2, tolerance=params.input.tolerance)
    )


def diagnose_multiple_sequences(sequences, params):
    """
    Print a diff between sequences.
    """
    logger.info("")
    for i in range(1, len(sequences)):
        logger.info("=" * 80)
        logger.info("Diff between sequence %d and %d", i - 1, i)
        logger.info("")
        print_sequence_diff(sequences[i - 1], sequences[i], params)
    logger.info("=" * 80)
    logger.info("")


def write_experiments(experiments, params):
    """
    Output the experiments to file.
    """
    if params.output.experiments:
        logger.info("-" * 80)
        logger.info("Writing experiments to %s", params.output.experiments)
        experiments.as_file(params.output.experiments, compact=params.output.compact)


def assert_single_sequence(experiments, params):
    """
    Print an error message if more than 1 sequence
    """
    sequences = [
        e.imageset for e in experiments if isinstance(e.imageset, ImageSequence)
    ]

    if len(sequences) > 1:

        # Print some info about multiple sequences
        diagnose_multiple_sequences(sequences, params)

        # Raise exception
        raise Sorry(
            """
    More than 1 sequence was found. Two things may be happening here:

    1. There really is more than 1 sequence. If you expected this to be the
        case, set the parameter allow_multiple_sequences=True. If you don't
        expect this, then check the input to dials.import.

    2. There may be something wrong with your image headers (for example,
        the rotation ranges of each image may not match up). You should
        investigate what went wrong, but you can force dials.import to treat
        your images as a single sequence by using the template=image_####.cbf
        parameter (see help).
    """
        )


def do_import(
    args: list[str] | None = None,
    *,
    phil: libtbx.phil.scope,
    configure_logging: bool = False,
):
    # Create the option parser
    usage = "dials.import [options] /path/to/image/files"
    parser = ArgumentParser(
        usage=usage,
        sort_options=True,
        phil=phil,
        read_experiments_from_images=True,
        epilog=help_message,
    )

    # Parse the command line arguments in two passes
    params, options = parser.parse_args(
        args=args, show_diff_phil=False, quick_parse=True
    )

    if configure_logging:
        log.config(verbosity=options.verbose, logfile=params.output.log)

    logger.info(dials_version())

    # Parse the command line arguments completely
    if params.input.ignore_unhandled:
        params, options, unhandled = parser.parse_args(
            args=args, show_diff_phil=False, return_unhandled=True
        )
        # Remove any False values from unhandled (eliminate empty strings)
        unhandled = [x for x in unhandled if x]
    else:
        params, options = parser.parse_args(args=args, show_diff_phil=False)
        unhandled = None

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    # Print a warning if something unhandled
    if unhandled:
        msg = "Unable to handle the following arguments:\n"
        msg += "\n".join(["  %s" % a for a in unhandled])
        msg += "\n"
        logger.warning(msg)

    # Print help if no input
    if len(params.input.experiments) == 0 and not (
        params.input.template or params.input.directory
    ):
        parser.print_help()
        sys.exit(0)

    # Re-extract the imagesets to rebuild experiments from
    imagesets = _extract_or_read_imagesets(params)

    metadata_updater = MetaDataUpdater(params)
    experiments = metadata_updater(imagesets)

    # Compute some numbers
    num_sweeps = 0
    num_still_sequences = 0
    num_stills = 0
    num_images = 0

    # importing a lot of experiments all pointing at one imageset should
    # work gracefully
    counted_imagesets = []

    for e in experiments:
        if e.imageset in counted_imagesets:
            continue
        if isinstance(e.imageset, ImageSequence):
            if e.imageset.get_scan().is_still():
                num_still_sequences += 1
            else:
                num_sweeps += 1
        else:
            num_stills += 1
        num_images += len(e.imageset)
        counted_imagesets.append(e.imageset)

    format_list = {str(e.imageset.get_format_class()) for e in experiments}

    # Print out some bulk info
    logger.info("-" * 80)
    for f in format_list:
        logger.info("  format: %s", f)
    logger.info("  num images: %d", num_images)
    logger.info("  sequences:")
    logger.info("    still:    %d", num_still_sequences)
    logger.info("    sweep:    %d", num_sweeps)
    logger.info("  num stills: %d", num_stills)

    # Print out info for all experiments
    for experiment in experiments:

        # Print some experiment info - override the output of image range
        # if appropriate
        image_range = params.geometry.scan.image_range
        if isinstance(experiment.imageset, ImageSequence):
            imageset_type = "sequence"
        else:
            imageset_type = "stills"

        logger.debug("-" * 80)
        logger.debug("  format: %s", str(experiment.imageset.get_format_class()))
        logger.debug("  imageset type: %s", imageset_type)
        if image_range is None:
            logger.debug("  num images:    %d", len(experiment.imageset))
        else:
            logger.debug("  num images:    %d", image_range[1] - image_range[0] + 1)

        logger.debug("")
        logger.debug(experiment.imageset.get_beam())
        logger.debug(experiment.imageset.get_goniometer())
        logger.debug(experiment.imageset.get_detector())
        logger.debug(experiment.imageset.get_scan())

    # Only allow a single sequence
    if params.input.allow_multiple_sequences is False:
        assert_single_sequence(experiments, params)

    # Write the experiments to file
    write_experiments(experiments, params)
    return experiments


@show_mail_handle_errors()
def run(args=None, *, phil=phil_scope):
    do_import(args, phil=phil, configure_logging=True)


if __name__ == "__main__":
    run()



from __future__ import annotations

from cctbx.array_family import flex

import dials_algorithms_indexing_ext as ext
from dials.algorithms.indexing import DialsIndexError


class AssignIndicesStrategy:
    def __init__(self, d_min=None):
        self._d_min = d_min

    def __call__(self, reciprocal_lattice_vectors):
        raise NotImplementedError()


class AssignIndicesGlobal(AssignIndicesStrategy):
    def __init__(self, tolerance=0.3):
        super().__init__()
        self._tolerance = tolerance

    def __call__(self, reflections, experiments, d_min=None):
        reciprocal_lattice_points = reflections["rlp"]
        reflections["miller_index"] = flex.miller_index(len(reflections), (0, 0, 0))
        if d_min is not None:
            d_spacings = 1 / reciprocal_lattice_points.norms()
            inside_resolution_limit = d_spacings > d_min
        else:
            inside_resolution_limit = flex.bool(reciprocal_lattice_points.size(), True)
        sel = inside_resolution_limit & (reflections["id"] == -1)
        isel = sel.iselection()
        rlps = reciprocal_lattice_points.select(isel)
        refs = reflections.select(isel)
        phi = refs["xyzobs.mm.value"].parts()[2]

        UB_matrices = flex.mat3_double([cm.get_A() for cm in experiments.crystals()])
        imgset_ids = reflections["imageset_id"].select(sel)

        for i_imgset, imgset in enumerate(experiments.imagesets()):
            sel_imgset = imgset_ids == i_imgset

            result = ext.AssignIndices(
                rlps.select(sel_imgset),
                phi.select(sel_imgset),
                UB_matrices,
                tolerance=self._tolerance,
            )

            miller_indices = result.miller_indices()
            crystal_ids = result.crystal_ids()

            expt_ids = flex.int(crystal_ids.size(), -1)
            for i_cryst, cryst in enumerate(experiments.crystals()):
                sel_cryst = crystal_ids == i_cryst
                for i_expt in experiments.where(crystal=cryst, imageset=imgset):
                    expt_ids.set_selected(sel_cryst, i_expt)
                    if experiments[i_expt].identifier:
                        reflections.experiment_identifiers()[i_expt] = experiments[
                            i_expt
                        ].identifier

            reflections["miller_index"].set_selected(
                isel.select(sel_imgset), miller_indices
            )
            reflections["id"].set_selected(isel.select(sel_imgset), expt_ids)
            reflections.set_flags(
                reflections["miller_index"] != (0, 0, 0), reflections.flags.indexed
            )
            reflections["id"].set_selected(reflections["miller_index"] == (0, 0, 0), -1)


class AssignIndicesLocal(AssignIndicesStrategy):
    def __init__(
        self, d_min=None, epsilon=0.05, delta=8, l_min=0.8, nearest_neighbours=20
    ):
        super().__init__()
        self._epsilon = epsilon
        self._delta = delta
        self._l_min = l_min
        self._nearest_neighbours = nearest_neighbours

    def __call__(self, reflections, experiments, d_min=None):
        from libtbx.math_utils import nearest_integer as nint
        from scitbx import matrix

        reciprocal_lattice_points = reflections["rlp"]
        if "miller_index" not in reflections:
            reflections["miller_index"] = flex.miller_index(len(reflections))
        if d_min is not None:
            d_spacings = 1 / reciprocal_lattice_points.norms()
            inside_resolution_limit = d_spacings > d_min
        else:
            inside_resolution_limit = flex.bool(reciprocal_lattice_points.size(), True)
        sel = inside_resolution_limit & (reflections["id"] == -1)
        isel = sel.iselection()
        rlps = reciprocal_lattice_points.select(isel)
        refs = reflections.select(isel)
        phi = refs["xyzobs.mm.value"].parts()[2]

        if len(rlps) <= self._nearest_neighbours:
            raise DialsIndexError(
                "index_assignment.local.nearest_neighbour must be smaller than the number of accepted reflections (%d)"
                % len(rlps)
            )

        UB_matrices = flex.mat3_double([cm.get_A() for cm in experiments.crystals()])

        result = ext.AssignIndicesLocal(
            rlps,
            phi,
            UB_matrices,
            epsilon=self._epsilon,
            delta=self._delta,
            l_min=self._l_min,
            nearest_neighbours=self._nearest_neighbours,
        )
        miller_indices = result.miller_indices()
        crystal_ids = result.crystal_ids()
        hkl = miller_indices.as_vec3_double().iround()

        assert miller_indices.select(crystal_ids < 0).all_eq((0, 0, 0))

        for i_cryst in set(crystal_ids):
            if i_cryst < 0:
                continue

            A = matrix.sqr(experiments[i_cryst].crystal.get_A())
            A_inv = A.inverse()

            cryst_sel = crystal_ids == i_cryst
            rlp_sel = rlps.select(cryst_sel)
            hkl_sel = hkl.select(cryst_sel).as_vec3_double()

            d_sel = 1 / rlp_sel.norms()
            d_perm = flex.sort_permutation(d_sel, reverse=True)

            hf_0 = A_inv * rlp_sel[d_perm[0]]
            h_0 = matrix.col([nint(j) for j in hf_0.elems])
            offset = h_0 - matrix.col(hkl_sel[d_perm[0]])
            # print "offset:", offset.elems

            h = hkl_sel + flex.vec3_double(hkl_sel.size(), offset.elems)

            refs["miller_index"].set_selected(
                cryst_sel, flex.miller_index(list(h.iround()))
            )
            refs["id"].set_selected(cryst_sel, i_cryst)

        crystal_ids.set_selected(crystal_ids < 0, -1)
        refs["id"] = crystal_ids
        refs["miller_index"].set_selected(crystal_ids < 0, (0, 0, 0))

        reflections["miller_index"].set_selected(isel, refs["miller_index"])
        reflections["id"].set_selected(isel, refs["id"])
        reflections.set_flags(
            reflections["miller_index"] != (0, 0, 0), reflections.flags.indexed
        )



"""
A standalone program to explore error model minimisation after scaling.

Scale factors and outliers determined from scaling are respected.
Profile intensities will be used by default. Profile/summation intensity
combination can be used, with the options:
    intensity_choice=combine combine.Imid={value}

All scaling corrections are applied to the intensities and variances (except
of course the error model adjustment) before the analysis is done.
"""

from __future__ import annotations

import json
import logging
import sys

from jinja2 import ChoiceLoader, Environment, PackageLoader

import libtbx.phil
from dxtbx import flumpy
from iotbx import phil

from dials.algorithms.scaling.combine_intensities import combine_intensities
from dials.algorithms.scaling.error_model.engine import run_error_model_refinement
from dials.algorithms.scaling.error_model.error_model import (
    BasicErrorModel,
    calc_deltahl,
    calc_sigmaprime,
)
from dials.algorithms.scaling.error_model.error_model_target import (
    calculate_regression_x_y,
)
from dials.algorithms.scaling.Ih_table import IhTable
from dials.algorithms.scaling.plots import (
    error_model_variance_plot,
    error_regression_plot,
    normal_probability_plot,
)
from dials.algorithms.scaling.scaling_library import choose_initial_scaling_intensities
from dials.algorithms.scaling.scaling_utilities import calculate_prescaling_correction
from dials.report.plots import i_over_sig_i_vs_i_plot
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

try:
    from typing import List
except ImportError:
    pass

logger = logging.getLogger("dials")

phil_scope = phil.parse(
    """
    include scope dials.algorithms.scaling.error_model.error_model.phil_scope
    intensity_choice = *profile sum combine
        .type = choice
        .help = "Use profile or summation intensities"
    combine.Imid = None
        .type = float
        .help = "Midpoint value to use when combining profile/summation intensities"
    output {
        log = dials.refine_error_model.log
            .type = str
            .help = "The log filename"
        html = error_model.html
            .type = str
            .help = "Filename for html report of error model"
        json = None
            .type = str
            .help = "Filename for json export of html report data"
    }
    """,
    process_includes=True,
)


def refine_error_model(params, experiments, reflection_tables):
    """Do error model refinement."""

    # prepare relevant data for datastructures
    for i, table in enumerate(reflection_tables):
        # First get the good data
        table = table.select(~table.get_flags(table.flags.bad_for_scaling, all=False))

        # Now chose intensities, ideally these two options could be combined
        # with a smart refactor
        if params.intensity_choice == "combine":
            if not params.combine.Imid:
                sys.exit("Imid value must be provided if intensity_choice=combine")
            table = calculate_prescaling_correction(table)  # needed for below.
            I, V = combine_intensities(table, params.combine.Imid)
            table["intensity"] = I
            table["variance"] = V
        else:
            table = choose_initial_scaling_intensities(
                table, intensity_choice=params.intensity_choice
            )
        reflection_tables[i] = table
    space_group = experiments[0].crystal.get_space_group()
    Ih_table = IhTable(
        reflection_tables, space_group, additional_cols=["partiality"], anomalous=True
    )

    # now do the error model refinement
    model = BasicErrorModel(basic_params=params.basic)
    try:
        model = run_error_model_refinement(model, Ih_table)
    except (ValueError, RuntimeError) as e:
        logger.info(e)
    else:
        return model


def make_output(model, params):
    """Get relevant data from the model and make html report."""

    if not (params.output.html or params.output.json):
        return

    data = {}
    table = model.filtered_Ih_table
    data["intensity"] = flumpy.from_numpy(table.intensities)
    sigmaprime = calc_sigmaprime(model.parameters, table)
    data["delta_hl"] = calc_deltahl(table, table.calc_nh(), sigmaprime)
    data["inv_scale"] = table.inverse_scale_factors
    data["sigma"] = flumpy.from_numpy(sigmaprime * table.inverse_scale_factors)
    data["binning_info"] = model.binner.binning_info
    d = {"error_model_plots": {}}
    d["error_model_plots"].update(normal_probability_plot(data))
    d["error_model_plots"].update(
        i_over_sig_i_vs_i_plot(data["intensity"], data["sigma"])
    )
    d["error_model_plots"].update(error_model_variance_plot(data))

    if params.basic.minimisation == "regression":
        x, y = calculate_regression_x_y(model.filtered_Ih_table)
        data["regression_x"] = x
        data["regression_y"] = y
        data["model_a"] = model.parameters[0]
        data["model_b"] = model.parameters[1]
        d["error_model_plots"].update(error_regression_plot(data))

    if params.output.html:
        logger.info("Writing html report to: %s", params.output.html)
        loader = ChoiceLoader(
            [
                PackageLoader("dials", "templates"),
                PackageLoader("dials", "static", encoding="utf-8"),
            ]
        )
        env = Environment(loader=loader)
        template = env.get_template("simple_report.html")
        html = template.render(
            page_title="DIALS error model refinement report",
            panel_title="Error distribution plots",
            panel_id="error_model_plots",
            graphs=d["error_model_plots"],
        )
        with open(params.output.html, "wb") as f:
            f.write(html.encode("utf-8", "xmlcharrefreplace"))

    if params.output.json:
        logger.info("Writing html report data to: %s", params.output.json)
        with open(params.output.json, "w") as outfile:
            json.dump(d, outfile)


@show_mail_handle_errors()
def run(args: List[str] = None, phil: libtbx.phil.scope = phil_scope) -> None:
    """Run the scaling from the command-line."""
    usage = """Usage: dials.refine_error_model scaled.refl scaled.expt [options]"""

    parser = ArgumentParser(
        usage=usage,
        read_experiments=True,
        read_reflections=True,
        phil=phil,
        check_format=False,
        epilog=__doc__,
    )
    params, options = parser.parse_args(args=args, show_diff_phil=False)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    log.config(verbosity=options.verbose, logfile=params.output.log)
    logger.info(dials_version())

    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    model = refine_error_model(params, experiments, reflections)

    if model:
        make_output(model, params)
    logger.info("Finished")


if __name__ == "__main__":
    run()



"""ΔCC½ algorithm definitions"""


from __future__ import annotations

import logging
from collections import defaultdict
from math import sqrt

from jinja2 import ChoiceLoader, Environment, PackageLoader

from cctbx import uctbx
from iotbx import mtz

from dials.algorithms.scaling.scale_and_filter import (
    make_histogram_plots,
    make_per_dataset_plot,
)
from dials.algorithms.statistics.delta_cchalf import PerGroupCChalfStatistics
from dials.array_family import flex
from dials.util.exclude_images import exclude_image_ranges_for_scaling
from dials.util.filter_reflections import filter_reflection_table
from dials.util.multi_dataset_handling import select_datasets_on_identifiers

logger = logging.getLogger("dials.command_line.compute_delta_cchalf")


class CCHalfFromMTZ:

    """
    Run a cc-half algorithm using an MTZ file.
    """

    def __init__(self, params, filename):
        self.params = params

        table, unit_cell, space_group = self.read_mtzfile(
            filename, self.params.mtz.batch_offset
        )
        table["group"] = table["dataset"]

        self.algorithm = DeltaCCHalf(table, unit_cell, space_group, params)
        self.results_summary = {}

    def run(self):
        self.algorithm.run()
        # update results

        # now do the exclusion
        results = self.algorithm.results_summary
        cc_half_values = results["per_dataset_delta_cc_half_values"]
        below_cutoff = (
            flex.double(cc_half_values["delta_cc_half_values"])
            < results["dataset_removal"]["cutoff_value"]
        )
        ids_to_remove = flex.int(cc_half_values["datasets"]).select(below_cutoff)
        for id_ in sorted(ids_to_remove):
            logger.info("Dataset %d is below the cutoff", id_)
        results["dataset_removal"].update({"experiments_fully_removed": ids_to_remove})
        self.results_summary = results

    def output(self):
        self.algorithm.output_table()
        self.algorithm.output_html_report()

    @staticmethod
    def read_mtzfile(filename, batch_offset=None):
        """
        Read the mtz file
        """
        miller_arrays = mtz.object(file_name=filename).as_miller_arrays(
            merge_equivalents=False
        )

        # Select the desired columns
        intensities = None
        batches = None
        for array in miller_arrays:
            if array.info().labels == ["I", "SIGI"]:
                intensities = array
            if array.info().labels == ["BATCH"]:
                batches = array
        if not intensities:
            raise KeyError("Intensities not found in mtz file, expected labels I, SIGI")
        if not batches:
            raise KeyError("Batch values not found")
        if batches.data().size() != intensities.data().size():
            raise ValueError("Batch and intensity array sizes do not match")

        # Get the unit cell and space group
        unit_cell = intensities.unit_cell()
        space_group = intensities.crystal_symmetry().space_group()

        # The reflection data
        table = flex.reflection_table()
        table["miller_index"] = intensities.indices()
        table["intensity"] = intensities.data()
        table["variance"] = flex.pow2(intensities.sigmas())

        # Create unit cell list
        zeroed_batches = batches.data() - flex.min(batches.data())
        dataset = flex.int(table.size(), 0)
        sorted_batches = flex.sorted(zeroed_batches)
        sel_perm = flex.sort_permutation(zeroed_batches)

        if not batch_offset:
            previous = 0
            potential_batch_offsets = flex.double()
            for i, b in enumerate(sorted_batches):
                if b - previous > 1:
                    potential_batch_offsets.append(b - previous)
                previous = b
            potential = flex.sorted(potential_batch_offsets)
            # potential is a list of low numbers (where images may not have any spots)
            # and larger numbers between batches.
            if len(potential) == 1:
                batch_offset = potential[0]
                logger.info(
                    """
Using a batch offset of %s to split datasets.
Batch offset can be specified with mtz.batch_offset=
""",
                    batch_offset,
                )
            elif len(potential) > 1:
                diffs = flex.double(
                    [potential[i + 1] - p for i, p in enumerate(potential[:-1])]
                )
                i = flex.sort_permutation(diffs)[-1]
                batch_offset = int(potential[i + 1] - (0.2 * diffs[i]))
                logger.info(
                    """
Using an approximate batch offset of %s to split datasets.
Batch offset can be specified with mtz.batch_offset=
""",
                    batch_offset,
                )
            else:
                batch_offset = 1

        previous = 0
        dataset_no = 0
        for i, b in enumerate(sorted_batches):
            if b - previous > batch_offset - 1:
                dataset_no += 1
            dataset[i] = dataset_no
            previous = b

        table["dataset"] = flex.int(table.size(), 0)
        table["dataset"].set_selected(sel_perm, dataset)

        return table, unit_cell, space_group


class CCHalfFromDials:

    """
    Run a cc-half algorithm using dials datafiles.
    """

    def __init__(self, params, experiments, reflection_table):
        self.params = params
        self.experiments = experiments
        self.reflection_table = reflection_table
        self.filtered_reflection_table = None
        # prepare data
        self.results_summary = {}
        self.group_to_datasetid_and_range = {}
        self.datasetid_to_groups = defaultdict(list)

        table, unit_cell, space_group = self.read_experiments(
            self.experiments, self.reflection_table
        )

        if self.params.mode == "dataset":
            table["group"] = table["dataset"]
        elif self.params.mode == "image_group":
            # set up tracking of groups to expts

            image_groups = flex.int(table["dataset"].size(), 0)
            counter = 0
            for id_ in set(table["dataset"]):
                sel = table["dataset"] == id_
                expid = table.experiment_identifiers()[id_]
                images_in_dataset = table["image"].select(sel)
                unique_images = set(images_in_dataset)
                min_img, max_img = (min(unique_images), max(unique_images))
                group_starts = list(range(min_img, max_img + 1, self.params.group_size))
                if len(group_starts) > 1:
                    if max_img - group_starts[-1] < (self.params.group_size - 1):
                        del group_starts[-1]

                for i, start in enumerate(group_starts[:-1]):
                    group_sel = (images_in_dataset >= start) & (
                        images_in_dataset < group_starts[i + 1]
                    )
                    image_groups.set_selected(
                        (sel.iselection().select(group_sel)), counter
                    )
                    self.group_to_datasetid_and_range[counter] = (
                        expid,
                        (start, group_starts[i + 1] - 1),
                    )
                    self.datasetid_to_groups[expid].append(counter)
                    counter += 1
                # now do last group
                group_sel = images_in_dataset >= group_starts[-1]
                image_groups.set_selected((sel.iselection().select(group_sel)), counter)
                self.group_to_datasetid_and_range[counter] = (
                    expid,
                    (group_starts[-1], max_img),
                )
                self.datasetid_to_groups[expid].append(counter)
                counter += 1

            table["group"] = image_groups

        self.algorithm = DeltaCCHalf(table, unit_cell, space_group, params)

    def run(self):
        """Run the ΔCC½ algorithm and then exclude data as appropriate"""
        self.algorithm.run()

        # now do the exclusion
        results = self.algorithm.results_summary
        cc_half_values = results["per_dataset_delta_cc_half_values"]
        below_cutoff = (
            flex.double(cc_half_values["delta_cc_half_values"])
            < results["dataset_removal"]["cutoff_value"]
        )
        ids_to_remove = flex.int(cc_half_values["datasets"]).select(below_cutoff)

        if self.params.mode == "dataset":
            filtered_reflections = self.remove_datasets_below_cutoff(
                self.experiments, self.reflection_table, ids_to_remove, results
            )
        elif self.params.mode == "image_group":
            filtered_reflections = self.remove_image_ranges_below_cutoff(
                self.experiments,
                self.reflection_table,
                ids_to_remove,
                self.group_to_datasetid_and_range,
                self.datasetid_to_groups,
                results,
            )
        self.filtered_reflection_table = filtered_reflections
        self.results_summary = results

    def output(self):
        """Save the output data and updated datafiles."""
        self.algorithm.output_table()
        logger.info(
            "Saving %d reflections to %s",
            self.filtered_reflection_table.size(),
            self.params.output.reflections,
        )
        self.filtered_reflection_table.as_file(self.params.output.reflections)

        logger.info("Saving the experiments to %s", self.params.output.experiments)
        self.experiments.as_file(self.params.output.experiments)
        self.algorithm.output_html_report()

    @staticmethod
    def remove_datasets_below_cutoff(
        experiments, reflections, ids_to_remove, results_summary
    ):
        """Remove the datasets with ids in ids_to_remove.

        Remove from the experiemnts and reflections and add information to the
        results summary dict.

        Returns:
          output_reflections: The reflection table with data removed.
        """
        n_valid_reflections = reflections.get_flags(reflections.flags.scaled).count(
            True
        )

        datasets_to_remove = []
        ids_removed = []
        for id_ in sorted(ids_to_remove):
            logger.info("Removing dataset %d", id_)
            datasets_to_remove.append(reflections.experiment_identifiers()[id_])
            ids_removed.append(id_)
        output_reflections = reflections.remove_on_experiment_identifiers(
            datasets_to_remove
        )
        experiments.remove_on_experiment_identifiers(datasets_to_remove)
        output_reflections.assert_experiment_identifiers_are_consistent(experiments)

        n_valid_filtered_reflections = output_reflections.get_flags(
            output_reflections.flags.scaled
        ).count(True)
        results_summary["dataset_removal"].update(
            {
                "experiments_fully_removed": datasets_to_remove,
                "experiment_ids_fully_removed": ids_removed,
                "n_reflections_removed": n_valid_reflections
                - n_valid_filtered_reflections,
            }
        )
        return output_reflections

    @staticmethod
    def remove_image_ranges_below_cutoff(
        experiments,
        reflections,
        ids_to_remove,
        image_group_to_expid_and_range,
        expid_to_image_groups,
        results_summary,
    ):
        """Remove image ranges from the datasets."""
        n_valid_reflections = reflections.get_flags(reflections.flags.scaled).count(
            True
        )
        expid_to_tableid = {
            v: k
            for k, v in zip(
                reflections.experiment_identifiers().keys(),
                reflections.experiment_identifiers().values(),
            )
        }

        experiments_to_delete = []
        exclude_images = []
        image_ranges_removed = []  # track for results summary
        n_removed_this_cycle = 1
        while n_removed_this_cycle != 0:
            other_potential_ids_to_remove = []
            n_removed_this_cycle = 0
            for id_ in sorted(ids_to_remove):
                exp_id, image_range = image_group_to_expid_and_range[id_]  # identifier
                if (
                    expid_to_image_groups[exp_id][-1] == id_
                    or expid_to_image_groups[exp_id][0] == id_
                ):  # is at edge of scan.
                    # loc = list(experiments.identifiers()).index(exp_id)
                    table_id = expid_to_tableid[exp_id]
                    image_ranges_removed.append([image_range, table_id])
                    logger.info(
                        "Removing image range %s from experiment %s",
                        image_range,
                        table_id,
                    )
                    exclude_images.append(
                        [f"{table_id}:{image_range[0]}:{image_range[1]}"]
                    )
                    if expid_to_image_groups[exp_id][-1] == id_:
                        del expid_to_image_groups[exp_id][-1]
                    else:
                        del expid_to_image_groups[exp_id][0]
                    n_removed_this_cycle += 1
                else:
                    other_potential_ids_to_remove.append(id_)
            ids_to_remove = other_potential_ids_to_remove
        for id_ in other_potential_ids_to_remove:
            exp_id, image_range = image_group_to_expid_and_range[id_]
            table_id = expid_to_tableid[exp_id]
            logger.info(
                """Image range %s from experiment %s is below the cutoff, but not at the edge of a sweep.""",
                image_range,
                table_id,
            )

        # Now remove individual batches
        if -1 in reflections["id"]:
            reflections = reflections.select(reflections["id"] != -1)
        reflection_list = reflections.split_by_experiment_id()
        reflection_list, experiments = exclude_image_ranges_for_scaling(
            reflection_list, experiments, exclude_images
        )

        # check if any image groups were all outliers and missed by the analysis
        # This catches an edge case where there is an image group full of
        # outliers, which gets filtered out before the analysis but should
        # be set as not a valid image range.
        exclude_images = []
        for exp in experiments:
            # if any of the image ranges are not in the sets tested, exclude them
            tested = []
            for exp_id, imgrange in image_group_to_expid_and_range.values():
                if exp_id == exp.identifier:
                    tested.extend(list(range(imgrange[0], imgrange[1] + 1)))
            for imgrange in exp.scan.get_valid_image_ranges(exp.identifier):
                if all([j not in tested for j in range(imgrange[0], imgrange[1] + 1)]):
                    table_id = expid_to_tableid[exp.identifier]
                    exclude_images.append([f"{table_id}:{imgrange[0]}:{imgrange[1]}"])
                    logger.info(
                        "Removing %s due to scaling outlier group.", exclude_images[-1]
                    )
        if exclude_images:
            reflection_list, experiments = exclude_image_ranges_for_scaling(
                reflection_list, experiments, exclude_images
            )

        # if a whole experiment has been excluded: need to remove it here
        ids_removed = []
        for exp, refl in zip(experiments, reflection_list):
            if not exp.scan.get_valid_image_ranges(
                exp.identifier
            ):  # if all removed above
                experiments_to_delete.append(exp.identifier)
                ids_removed.append(refl.experiment_identifiers().keys()[0])
        if experiments_to_delete:
            experiments, reflection_list = select_datasets_on_identifiers(
                experiments, reflection_list, exclude_datasets=experiments_to_delete
            )
        assert len(reflection_list) == len(experiments)

        output_reflections = flex.reflection_table()
        for r in reflection_list:
            output_reflections.extend(r)

        n_valid_filtered_reflections = output_reflections.get_flags(
            output_reflections.flags.scaled
        ).count(True)
        results_summary["dataset_removal"].update(
            {
                "image_ranges_removed": image_ranges_removed,
                "experiments_fully_removed": experiments_to_delete,
                "experiment_ids_fully_removed": ids_removed,
                "n_reflections_removed": n_valid_reflections
                - n_valid_filtered_reflections,
            }
        )
        return output_reflections

    @staticmethod
    def read_experiments(experiments, reflection_table):
        """
        Get information from experiments and reflections
        """

        # Get space group and unit cell
        space_group = experiments[0].crystal.get_space_group()
        sgno = space_group.type().number()
        unit_cell_list = []
        for e in experiments:
            assert sgno == e.crystal.get_space_group().type().number()
            unit_cell_list.append(e.crystal.get_unit_cell())

        if len(unit_cell_list) > 1:
            # calc mean unit cell
            mean_parameters = [0, 0, 0, 0, 0, 0]
            for uc in unit_cell_list:
                for i in range(6):
                    mean_parameters[i] += uc.parameters()[i]
            for i in range(6):
                mean_parameters[i] /= len(unit_cell_list)
            mean_unit_cell = uctbx.unit_cell(mean_parameters)
        else:
            mean_unit_cell = unit_cell_list[0]

        # Require a dials scaled experiments file.
        filtered_table = filter_reflection_table(reflection_table, ["scale"])
        filtered_table["intensity"] = filtered_table["intensity.scale.value"]
        filtered_table["variance"] = filtered_table["intensity.scale.variance"]
        filtered_table["dataset"] = filtered_table["id"]
        filtered_table["image"] = (
            flex.floor(filtered_table["xyzobs.px.value"].parts()[2]).iround() + 1
        )

        return filtered_table, mean_unit_cell, space_group


class DeltaCCHalf:

    """
    Implementation of a ΔCC½ algorithm.
    """

    def __init__(self, reflection_table, median_unit_cell, space_group, params):
        self.reflection_table = reflection_table
        self.params = params
        self.median_unit_cell = median_unit_cell
        self.space_group = space_group
        self.delta_cchalf_i = {}
        self.results_summary = {
            "dataset_removal": {
                "mode": self.params.mode,
                "stdcutoff": self.params.stdcutoff,
            }
        }
        if len(set(reflection_table["group"])) == 1:
            if self.params.mode == "dataset":
                raise ValueError(
                    """Cannot perform delta-cc-half analysis in dataset mode with only one dataset.
For image group based delta-cc-half analysis, use the option mode=image_group"""
                )
            else:
                raise ValueError(
                    """Cannot perform delta-cc-half analysis on a single image group.
Choose a suitable option for group_size to divide the dataset into multiple groups"""
                )

    def run(self):
        """Run the delta_cc_half algorithm."""

        statistics = PerGroupCChalfStatistics(
            self.reflection_table,
            self.median_unit_cell,
            self.space_group,
            self.params.dmin,
            self.params.dmax,
            self.params.nbins,
        )

        statistics.run()

        self.delta_cchalf_i = statistics.delta_cchalf_i()
        self.results_summary["mean_cc_half"] = statistics._cchalf_mean
        # Print out the datasets in order of ΔCC½
        self.sort_deltacchalf_values(self.delta_cchalf_i, self.results_summary)

        cutoff_value = self._calculate_cutoff_value(
            self.delta_cchalf_i, self.params.stdcutoff
        )
        self.results_summary["dataset_removal"].update({"cutoff_value": cutoff_value})

    def output_table(self):
        """Write a text file with delta cchalf value"""
        if self.params.output.table:
            logger.info("Writing table to %s", self.params.output.table)
            with open(self.params.output.table, "w") as outfile:
                for dataset, cchalf in zip(
                    self.results_summary["per_dataset_delta_cc_half_values"][
                        "datasets"
                    ],
                    self.results_summary["per_dataset_delta_cc_half_values"][
                        "delta_cc_half_values"
                    ],
                ):
                    outfile.write("%d %f\n" % (dataset, cchalf))

    @staticmethod
    def sort_deltacchalf_values(delta_cchalf_i, results_summary):
        """Return the sorted datasets and cchalf values.

        Also add the sorted lists to the results summary. Datasets are sorted
        from low to high based on deltacchalf values."""
        datasets = list(delta_cchalf_i.keys())
        sorted_index = sorted(
            range(len(datasets)), key=lambda x: delta_cchalf_i[datasets[x]]
        )

        # sorted by deltacchalf from low to high
        sorted_cc_half_values = flex.double([])
        sorted_datasets = flex.int([])
        for i in sorted_index:
            val = delta_cchalf_i[datasets[i]]
            logger.info("Dataset: %d, ΔCC½: %.3f", datasets[i], 100 * val)
            sorted_cc_half_values.append(val)
            sorted_datasets.append(datasets[i])

        results_summary["per_dataset_delta_cc_half_values"] = {
            "datasets": list(sorted_datasets),
            "delta_cc_half_values": list(sorted_cc_half_values),
        }

        return sorted_datasets, sorted_cc_half_values

    @staticmethod
    def _calculate_cutoff_value(delta_cchalf_i, stdcutoff):
        Y = list(delta_cchalf_i.values())
        mean = sum(Y) / len(Y)
        sdev = sqrt(sum((yy - mean) ** 2 for yy in Y) / len(Y))
        logger.info(f"\nmean delta_cc_half: {(mean * 100):.3f}")
        logger.info(f"stddev delta_cc_half: {(sdev * 100):.3f}")
        cutoff_value = mean - stdcutoff * sdev
        logger.info(f"cutoff value: {(cutoff_value * 100):.3f} \n")
        return cutoff_value

    def output_html_report(self):
        if self.params.output.html:
            data = {"cc_half_plots": {}}
            res = {
                "delta_cc_half_values": self.results_summary[
                    "per_dataset_delta_cc_half_values"
                ]["delta_cc_half_values"],
                "mean_cc_half": self.results_summary["mean_cc_half"],
            }
            if "image_ranges_removed" in self.results_summary["dataset_removal"]:
                res["image_ranges_removed"] = self.results_summary["dataset_removal"][
                    "image_ranges_removed"
                ]
            else:
                res["removed_datasets"] = self.results_summary["dataset_removal"][
                    "experiments_fully_removed"
                ]
            data["cc_half_plots"].update(make_histogram_plots([res]))
            del data["cc_half_plots"]["mean_cc_one_half_vs_cycle"]
            data["cc_half_plots"].update(make_per_dataset_plot(self.delta_cchalf_i))

            logger.info("Writing html report to: %s", self.params.output.html)
            loader = ChoiceLoader(
                [
                    PackageLoader("dials", "templates"),
                    PackageLoader("dials", "static", encoding="utf-8"),
                ]
            )
            env = Environment(loader=loader)
            template = env.get_template("simple_report.html")
            html = template.render(
                page_title="ΔCC½ report",
                panel_title="Delta CC-Half plots",
                panel_id="cc_half_plots",
                graphs=data["cc_half_plots"],
            )
            with open(self.params.output.html, "wb") as f:
                f.write(html.encode("utf-8", "xmlcharrefreplace"))
